//
// Created by benedikt on 4/11/19.
//

#ifndef UMC_NEW_CYCLE_H
#define UMC_NEW_CYCLE_H

#include <unordered_set>
#include <boost/graph/graph_traits.hpp>

namespace UMC {
  template<typename Graph>
  class Cycle {
      friend std::hash<UMC::Cycle<Graph>>;
    public:
      typedef typename Graph::EdgeId Edge;

      explicit Cycle(const Graph* graph, std::unordered_set<Edge> edges, bool* bug = nullptr);

      const std::unordered_set<typename Graph::vertex_descriptor>& get_complement_vertices() const {
        return complement_vertices_;
      }

      const std::vector<bool>& get_complement_indicator() const {
        return complement_indicator_;
      }

      const std::unordered_set<Edge>& get_edges() const { return edges_; }

      virtual void write(std::ostream& os) const;

      virtual ~Cycle() = default;;

      bool operator==(const Cycle<Graph>& u) const;

    protected:
      const Graph* graph_;
    private:
      std::unordered_set<Edge> edges_;
      std::unordered_set<typename Graph::vertex_descriptor> complement_vertices_;
      std::vector<bool> complement_indicator_;
      std::vector<bool> edge_indicator_;

      template<class G>
      friend std::ostream& operator<<(std::ostream& os, Cycle<G> const& m);

      bool is_valid() const;
  };

  template<typename Graph>
  std::ostream& operator<<(std::ostream& os, Cycle<Graph> const& c) {
    c.write(os);
    return os;
  }

  template<typename Graph>
  class FECycle : public Cycle<Graph> {
    public:
      typedef typename Graph::EdgeId Edge;

      FECycle(const Cycle<Graph>& c, Edge fixed_edge) : Cycle<Graph>(c), fixed_edge(fixed_edge) {}

      virtual void write(std::ostream& os) const;

      virtual ~FECycle() = default;;
    private:
      Edge fixed_edge;
  };

  template<typename Graph>
  void FECycle<Graph>::write(std::ostream& os) const {
    Cycle<Graph>::write(os);
    auto e = this->graph_->get_edge(fixed_edge);
    auto v1 = boost::source(e, *this->graph_);
    auto v2 = boost::target(e, *this->graph_);
    os << " fixed edge: " << v1 << "<->" << v2;
  }

  template<typename Graph>
  bool Cycle<Graph>::is_valid() const {
    //check if cycle is really a cycle
    auto n = boost::num_vertices(*graph_);
    std::vector<bool> visited(n, false);
    size_t count = 0;
    if (edges_.size() == 0) {
      return false;
    }
    for (size_t i = 0; i < n; i++) {
      if (!complement_indicator_[i]) {
        typename Graph::vertex_descriptor current = i;
        auto edge_index = get(boost::edge_index, *graph_);
        bool first = true;
        auto previous = current;
        while (!visited[current]) {
          bool found = false;
          visited[current] = true;
          count++;
          typename Graph::vertex_descriptor t = -1;
          for (auto e: boost::make_iterator_range(boost::out_edges(current, *graph_))) {
            auto id = get(edge_index, e);
            if (edge_indicator_[id]) {
              auto target = boost::target(e, *graph_);
              assert(target != current);
              if (target != previous) {
                if (found) {
                  return false;
                }
                if (first) {
                  first = false;
                } else {
                  t = target;
                  found = true;
                }
              }
            }
          }
          if (!found) {
            return false;
          }
          previous = current;
          current = t;
        }
        if (current != i || count != edges_.size()) {
          return false;
        }
        break;
      }
    }
    return true;
  }

  template<typename Graph>
  Cycle<Graph>::Cycle(const Graph* graph, std::unordered_set<Edge> edges, bool* bug)
      : graph_(graph), edges_(edges), complement_indicator_(std::vector<bool>()), edge_indicator_(std::vector<bool>()) {
    auto n = boost::num_vertices(*graph_);
    auto m = boost::num_edges(*graph_);
    complement_indicator_.resize(n, true);
    edge_indicator_.resize(m, false);
    for (auto e : edges_) {
      auto ed = graph_->get_edge(e);
      auto v1 = boost::source(ed, *graph_);
      auto v2 = boost::target(ed, *graph_);
      complement_indicator_[v1] = false;
      complement_indicator_[v2] = false;
      edge_indicator_[e] = true;
    }
    for (size_t i = 0; i < n; i++) {
      if (complement_indicator_[i]) {
        complement_vertices_.insert(i);
      }
    }

    bool is_valid = this->is_valid();
    if (bug != nullptr) {
      *bug = !is_valid;
    } else if (!is_valid) {
      std::cerr << "Unexpected Wrong Cycle Error!" << std::endl;
      exit(1);
    }
  }

  template<typename Graph>
  void Cycle<Graph>::write(std::ostream& os) const {
    //very inefficient, but ok for our purposes
    auto first_edge = graph_->get_edge(*edges_.begin());
    auto begin = boost::source(first_edge, *graph_);
    auto last = begin;
    auto current = boost::target(first_edge, *graph_);
    os << begin << " -> " << current;
    while (current != begin) {
      for (auto e_id: edges_) {
        auto e = graph_->get_edge(e_id);
        auto v1 = boost::source(e, *graph_);
        auto v2 = boost::target(e, *graph_);
        if (v2 == current) {
          auto tmp = v2;
          v2 = v1;
          v1 = tmp;
        }
        if (v1 == current && v2 != last) {
          last = current;
          current = v2;
          os << " -> " << current;
          break;
        }
      }
    }
  }

  template<typename Graph>
  bool Cycle<Graph>::operator==(const Cycle<Graph>& u) const {
    return edges_ == u.edges_;
  }
}

namespace std {
  template<typename Graph>
  class hash<UMC::Cycle<Graph>> {
    public:
      size_t operator()(const UMC::Cycle<Graph>& c) const {
        return hash<std::vector<bool>>()(c.edge_indicator_);
      }
  };
}


#endif //UMC_NEW_CYCLE_H
