/*
 * CycleDatabase.h
 *
 *  Created on: Jul 28, 2015
 *      Author: klocker
 */

#ifndef SRC_CYCLEDATABASE_H_
#define SRC_CYCLEDATABASE_H_

#include <list>
#include <unordered_map>
#include <utility>
#include <vector>
#include <functional>

#include "Cycle.h"
#include "Hasse.h"

namespace UMC {
  template<typename _Tp>
  struct pointer_equal_to : public std::binary_function<const _Tp*, const _Tp*, bool> {
      _GLIBCXX14_CONSTEXPR
      bool
      operator()(_Tp const* const& __x, _Tp const* const& __y) const { return *__x == *__y; }
  };

  template<typename Graph>
  class CycleDatabase {
    private:
      typedef typename Graph::vertex_descriptor V;
      struct ComplementVerticesHasher;

      struct ComplementVertices {
          explicit ComplementVertices(const std::unordered_set<V>* vertices, const std::vector<bool>* indicator)
              : vs(vertices), indicator(indicator), hash(0) {
            hash = std::hash<std::vector<bool>>()(*indicator);
          }

          const std::unordered_set<V>* vs;
          const std::vector<bool>* indicator;
          std::size_t hash;

          bool operator==(const ComplementVertices& u) const;

          bool operator<(const ComplementVertices& u) const;
      };

      struct ComplementVerticesHasher {
          std::size_t operator()(const ComplementVertices& c) const {
            return c.hash;
          }
      };

      struct CycleHasher {
          std::size_t operator()(const Cycle<Graph>* const& c) const {
            return std::hash<Cycle<Graph>>()(*c);
          }
      };

    public:
      CycleDatabase(const Graph* g, const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges);

      virtual ~CycleDatabase();

      bool addCycle(const Cycle<Graph>& c);



      typedef std::unordered_set<const Cycle<Graph>*, CycleHasher, pointer_equal_to<Cycle<Graph>>> CycleSet;

      typedef std::unordered_map<typename Graph::EdgeId,
          std::unordered_map<ComplementVertices, CycleSet, ComplementVerticesHasher> >
          EdgeToCycleMap;

      const EdgeToCycleMap& getEdgeToCycleMap() const {
        return edgeToCycleMap_;
      }

      typedef std::unordered_map<ComplementVertices, std::pair<CycleSet,
          std::unordered_map<typename Graph::EdgeId, size_t>>, ComplementVerticesHasher> CyclesMap;

      const CyclesMap& getCycles() const {
        return cycles_;
      }

      const std::unordered_map<typename Graph::EdgeId, size_t>* getCountsOfCycle(const Cycle<Graph>& c) const;

    private:
      CycleDatabase(const Graph* g, const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges,
                    ComplementVertices root);

      void insert_empty_cycle(ComplementVertices vs);

      void removeIfNotNeeded(const ComplementVertices& vertexSet);

      const Graph* g_;
      CyclesMap cycles_;
      EdgeToCycleMap edgeToCycleMap_;
      typedef Hasse<ComplementVertices, ComplementVerticesHasher> CycleHasse;
      CycleHasse cycleHasse_;
      ComplementVertices root;

      void edge_is_dominated(const typename CycleHasse::Entry& entry, typename Graph::EdgeId edge,
                             std::unordered_set<ComplementVertices*>& modified_complements) {
        assert(cycles_.at(entry.val).second.at(edge) != 0);
        for (auto& child: entry.children) {
          auto& c = cycles_.at(child->val).second.at(edge);
          if (c == 0) {
            c = (size_t) -1;
            modified_complements.insert(&child->val);
            edge_is_dominated(*child, edge, modified_complements);
          } else if (c != (size_t) -1) {
            c = (size_t) -1;
            modified_complements.insert(&child->val);
            edgeToCycleMap_.at(edge).erase(child->val);
          }
        }
      }
  };

  template<typename Graph>
  CycleDatabase<Graph>::CycleDatabase(const Graph* g,
                                      const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges) :
      CycleDatabase(g, possibleFixedEdges,
                    ComplementVertices(new std::unordered_set<V>(),
                                       new std::vector<bool>(boost::num_vertices(*g), false)
                    )) {}

  template<typename Graph>
  CycleDatabase<Graph>::CycleDatabase(const Graph* g,
                                      const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges,
                                      ComplementVertices root) :
      g_(g), cycles_(CyclesMap()), edgeToCycleMap_(EdgeToCycleMap()),
      cycleHasse_(Hasse<ComplementVertices, ComplementVerticesHasher>(root)), root(root) {
    for (auto e : possibleFixedEdges) {
      edgeToCycleMap_[e];
    }
    insert_empty_cycle(root);
  }

  template<typename Graph>
  CycleDatabase<Graph>::~CycleDatabase() {
    for (auto& p: cycles_) {
      for (auto c: p.second.first) {
        delete c;
      }
    }
    delete root.vs;
    delete root.indicator;
  }

  template<typename Graph>
  void CycleDatabase<Graph>::insert_empty_cycle(ComplementVertices vs) {
    auto map = std::unordered_map<typename Graph::EdgeId, size_t>();
    for (auto p : edgeToCycleMap_) {
      map[p.first] = 0;
    }
    cycles_.insert(std::make_pair(
        vs,
        std::make_pair(CycleSet(), map)
    ));
  }

  template<typename Graph>
  const std::unordered_map<typename Graph::EdgeId, size_t>*
  CycleDatabase<Graph>::getCountsOfCycle(const Cycle<Graph>& c) const {
    auto& complement_vertices = c.get_complement_vertices();
    auto& complement_indicator = c.get_complement_indicator();
    auto complement = ComplementVertices(&complement_vertices, &complement_indicator);
    if (cycles_.count(complement) == 0) {
      return nullptr;
    }
    return &cycles_.at(complement).second;
  }

  template<typename Graph>
  bool CycleDatabase<Graph>::addCycle(const Cycle<Graph>& c) {
    auto& complement_vertices = c.get_complement_vertices();
    auto& complement_indicator = c.get_complement_indicator();
    auto complement = ComplementVertices(&complement_vertices, &complement_indicator);
    bool is_relevant = false;
    Output::local.cycles_found++;
    Output::global.cycles_found++;
    if (cycles_.count(complement) > 0) {
      auto& p = cycles_.at(complement);
      if (cycles_.at(complement).first.count(&c) > 0) { //cycle already in data structure
        return false;
      }
      for (auto edge : c.get_edges()) {
        if (p.second.count(edge) > 0 && (p.second.at(edge) == 0 || p.second.at(edge) == 1)) {
          is_relevant = true;
        }
      }
      if (!is_relevant) {
        return false;
      }
    }

    auto* cycle = new Cycle<Graph>(c);
    complement = ComplementVertices(&cycle->get_complement_vertices(), &cycle->get_complement_indicator());
    if (cycles_.count(complement) == 0) {
      insert_empty_cycle(complement);
    }
    auto& pair = cycles_.at(complement);
    //Globals::gatheredCycles++;

    for (auto edge : c.get_edges()) {
      if (pair.second.count(edge) > 0 && pair.second.count(edge) > 0 && pair.second.at(edge) != (size_t) -1) {
        ++pair.second.at(edge);
        if (pair.second.at(edge) <= 2) {
          edgeToCycleMap_.at(edge)[complement].insert(cycle);
        }
      }
    }

    //bool newAndHasChildren = false;
    cycleHasse_.add(complement);
    for (auto& parent: cycleHasse_[complement].parents) {
      for (auto& p: cycles_[parent->val].second) {
        if (p.second != 0) {
          //either is dominated i.e. -1 or has a count > 0
          if (pair.second.count(p.first) > 0 && pair.second.at(p.first) != (size_t) -1) {
            edgeToCycleMap_.at(p.first).erase(complement);
            pair.second.at(p.first) = (size_t) -1;
          }
        }
      }
    }

    auto num_vertex_complements = cycleHasse_.numEntries();
    if (num_vertex_complements > Output::local.max_num_vertex_complements) {
      Output::local.max_num_vertex_complements = num_vertex_complements;
      if (num_vertex_complements > Output::global.max_num_vertex_complements) {
        Output::global.max_num_vertex_complements = num_vertex_complements;
      }
    }

    is_relevant = false;
    std::unordered_set<ComplementVertices*> modified_complements;
    for (auto edge : c.get_edges()) {
      if (pair.second.count(edge) > 0) {
        auto count = pair.second.at(edge);
        if (count == 1) {
          is_relevant = true;
          //new edge for this vertex set => set dominated for all undominated children
          edge_is_dominated(cycleHasse_[complement], edge, modified_complements);
        } else if (count == 2) {
          is_relevant = true; //still relevant
        }
      }
    }

    for (auto& complement2: modified_complements) {
      removeIfNotNeeded(*complement2);
    }

    if (!is_relevant) { //should not happen in ILP approaches
      cycleHasse_.erase(complement);
    }

    Output::local.cycles_inserted++;
    Output::global.cycles_inserted++;
    pair.first.insert(cycle);
    return true;
  }

  template<typename Graph>
  void CycleDatabase<Graph>::removeIfNotNeeded(const ComplementVertices& vertexSet) {
    for (auto& p : cycles_[vertexSet].second) {
      if (p.second != (size_t) -1 && p.second != 0) {
        return; //is needed do nothing
      }
    }

    //remove all corresponding cycles
    Output::local.cycles_removed += cycles_[vertexSet].first.size();
    Output::global.cycles_removed += cycles_[vertexSet].first.size();
    //Globals::removedCycles += cycles_[vertexSet].first.size();
    cycleHasse_.erase(vertexSet);
  }

  template<typename Graph>
  bool CycleDatabase<Graph>::ComplementVertices::operator==(const CycleDatabase<Graph>::ComplementVertices& u) const {
    return u.vs == vs || *u.vs == *vs;
  }

  template<typename Graph>
  bool CycleDatabase<Graph>::ComplementVertices::operator<(const CycleDatabase<Graph>::ComplementVertices& u) const {
    if (vs->size() >= u.vs->size()) {
      return false;
    }
    for (auto& p: *vs) {
      if (!(*u.indicator).at(p)) {
        return false;
      }
    }
    return true;
  }

} /* namespace UMC */


#endif /* SRC_CYCLEDATABASE_H_ */
