//
// Created by benedikt on 6/16/19.
//

#ifndef UMC_NEW_ENUMERATIONSOLVER_H
#define UMC_NEW_ENUMERATIONSOLVER_H

#include "UMCAbstract.h"
#include <boost/graph/graph_traits.hpp>
#include <boost/range/iterator_range.hpp>
#include <boost/graph/adjacency_list.hpp>

namespace UMC {
  namespace EnumerationInternal {
    enum VertexState {
        Unvisited,
        Visited,
        End,
        Fixed,
        Dropped
    };
  }
  template<typename G>
  class EnumerationSolver : public UMCAbstract<G> {
    public:
      explicit EnumerationSolver(bool only_dominating = false, bool minimal = false)
          : UMCAbstract<G>(only_dominating, minimal) {}

    protected:
      FECycle <G>* searchCycles(const G& g, const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) override {
        this->searchCyclesByFixingVertices(g, possibleFixedEdges);
        return nullptr;
      }

      void searchCyclesWithFixedVertex(const G& g, typename G::vertex_descriptor fixedVertex,
                                       std::unordered_set<typename G::vertex_descriptor>& forbidden_vertices,
                                       const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) override;

    private:
      //TODO: if bottleneck do not remove edges but only mark them (or alternatively use the vertexStates to find out
      // if edge is still useable) and then use an own degree counter per vertex
      std::vector<std::vector<std::pair<typename G::vertex_descriptor, typename G::EdgeId>>> adjecency_list;
      std::vector<size_t> degree;
      std::vector<typename G::vertex_descriptor> ends;
      std::vector<EnumerationInternal::VertexState> vertexStates;
      int unsatisfied_edges;
      std::unordered_set<typename G::EdgeId> cycle;
      const G* _g;
      unsigned long int recursion_count;
      unsigned long int next_check;
      bool time_limit_reached = false;
      std::clock_t start;


      /*std::pair<typename G::vertex_descriptor, typename G::EdgeId> remove_edge(
          typename G::vertex_descriptor v, typename G::vertex_descriptor w, bool reverse = true);*/

      bool find_cycles_from_ends();

      bool check_degree(typename G::vertex_descriptor v) {
        size_t d = 0;
        for (auto& p: adjecency_list[v]) {
          auto next_end = p.first;
          if (vertexStates[next_end] != EnumerationInternal::VertexState::Visited &&
              vertexStates[next_end] != EnumerationInternal::VertexState::Dropped) {
            d++;
          }
        }
        return d == degree[v];
      }

      bool check_degrees() {
        for (typename G::vertex_descriptor v = 0; v < boost::num_vertices(*_g); v++) {
          if (vertexStates[v] != EnumerationInternal::VertexState::Visited &&
              vertexStates[v] != EnumerationInternal::VertexState::Dropped && !check_degree(v)) {
            return false;
          }
        }
        return true;
      }

      typename G::vertex_descriptor start_v1;
      typename G::vertex_descriptor start_v2;

      bool check_unsatisfied_edges() {
        int unsatisfied = 0;
        for (auto e: boost::make_iterator_range(boost::edges(*_g))) {
          auto v1 = boost::target(e, *_g);
          auto v2 = boost::source(e, *_g);
          if ((v1 != start_v1 || v2 != start_v2) && (v2 != start_v1 || v1 != start_v2) &&
              vertexStates[v1] != EnumerationInternal::VertexState::Visited &&
              vertexStates[v2] != EnumerationInternal::VertexState::Visited) {
            unsatisfied++;
          }
        }
        return unsatisfied == unsatisfied_edges;
      }

      bool cycle_is_dominating(Cycle <G>& c) {
        for (auto e: boost::make_iterator_range(boost::edges(*_g))) {
          auto v1 = boost::target(e, *_g);
          auto v2 = boost::source(e, *_g);
          if (c.get_complement_indicator()[v1] && c.get_complement_indicator()[v2]) {
            return false;
          }
        }
        return true;
      }
  };

  /*template<typename G>
  std::pair<typename G::vertex_descriptor, typename G::EdgeId>
  EnumerationSolver<G>::remove_edge(typename G::vertex_descriptor v, typename G::vertex_descriptor w, bool reverse) {
    assert(vertexStates[v] != EnumerationInternal::VertexState::Visited);
    std::pair<typename G::vertex_descriptor, typename G::EdgeId> res = std::make_pair(-1, -1);
    for (size_t i = 0; i < adjecency_list[v].size(); i++) {
      if (adjecency_list[v][i].first == w) {
        res = adjecency_list[v][i];
        assert(res.first != -1);
        adjecency_list[v].erase(adjecency_list[v].begin() + i);
        if (v == 5 && w == 1 && res.second == 4) {
          int debug = 1;
        }
        std::cout << "remove edge " << v << "->" << w << std::endl;
        break;
      }
    }
    if (reverse) {
      remove_edge(w, v, false);
    }
    return res;
  }*/

  template<typename G>
  void EnumerationSolver<G>::searchCyclesWithFixedVertex(const G& g, typename G::vertex_descriptor fixedVertex,
                                                         std::unordered_set<typename G::vertex_descriptor>& forbidden_vertices,
                                                         const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) {
    _g = &g;
    auto n = boost::num_vertices(g);
    //construct adjecency list to not use the clumpy graph library TODO: test the real difference of own adjecency
    // list vs graph library
    adjecency_list.resize(n);
    degree.resize(n);
    unsatisfied_edges = boost::num_edges(g);
    vertexStates.clear();
    vertexStates.resize(n, EnumerationInternal::VertexState::Unvisited);
    ends.resize(2);
    auto edge_index = get(boost::edge_index, g);
    for (typename G::vertex_descriptor v = 0; v < n; v++) {
      adjecency_list[v].clear();
      degree[v] = 0;
      for (auto e: boost::make_iterator_range(boost::out_edges(v, g))) {
        auto w = boost::target(e, g);
        assert(w != v);
        adjecency_list[v].push_back(std::make_pair(w, get(edge_index, e)));
        if (forbidden_vertices.count(w) == 0) {
          degree[v]++;
        }
      }
    }
    for (auto v: forbidden_vertices) {
      vertexStates[v] = EnumerationInternal::VertexState::Dropped;
      if (this->minimal) {
        for (auto w: boost::make_iterator_range(boost::adjacent_vertices(v, g))) {
          if (vertexStates[w] == EnumerationInternal::VertexState::Dropped) {
            //two neighboring dropped vertices => not dominating
            return;
          }
          vertexStates[w] = EnumerationInternal::VertexState::Fixed;
        }
      }
    }

    if (!this->minimal) {
      //we also have to consider if there are any degree 0 or degree 1 vertices at the moment which are not forbidden
      // vertices
      bool found = true;
      while (found) {
        found = false;
        for (typename G::vertex_descriptor v = 0; v < n; v++) {
          if (vertexStates[v] != EnumerationInternal::VertexState::Dropped) {
            assert(vertexStates[v] == EnumerationInternal::VertexState::Unvisited);
            if (degree[v] <= 1) {
              vertexStates[v] = EnumerationInternal::VertexState::Dropped;
              if (degree[v] == 1) {
                //remove degree of neighboring vertex
                for (size_t j = 0; j < adjecency_list[v].size(); j++) {
                  auto w = adjecency_list[v][j].first;
                  if (vertexStates[w] != EnumerationInternal::VertexState::Dropped) {
                    degree[w]--;
                    found = true;
                    assert(check_degrees());
                  }
                }
              }
            }
          }
        }
      }
    }

    if (this->minimal) {
      //fix all neighbors of large vertices
      for (typename G::vertex_descriptor v = 0; v < n; v++) {
        if (adjecency_list[v].size() > 3) { //v is large
          for (auto p: adjecency_list[v]) {
            if (vertexStates[p.first] == EnumerationInternal::VertexState::Dropped) {
              //neighbor of large vertex got dropped
              return;
            }
            vertexStates[p.first] = EnumerationInternal::VertexState::Fixed;
          }
        }
      }
    }

    if (vertexStates[fixedVertex] == EnumerationInternal::VertexState::Dropped) {
      //cannot find any cycles with this vertex
      return;
    }

    assert(check_degrees());

    vertexStates[fixedVertex] = EnumerationInternal::VertexState::End;
    start_v1 = fixedVertex;

    unsatisfied_edges -= 1;
    assert(unsatisfied_edges >= 0);
    degree[fixedVertex]--;
    ends[0] = fixedVertex;
    for (size_t i = 0; i < adjecency_list[fixedVertex].size(); i++) {
      if (std::clock() > this->max_end_time) {
        return;
      }
      ends[1] = adjecency_list[fixedVertex][i].first;
      start_v2 = ends[1];
      auto state = vertexStates[ends[1]];
      if (state == EnumerationInternal::VertexState::Unvisited || state == EnumerationInternal::VertexState::Fixed) {
        auto e = adjecency_list[fixedVertex][i].second;
        cycle.insert(e);
        vertexStates[ends[1]] = EnumerationInternal::VertexState::End;

        //remove edge between ends
        adjecency_list[fixedVertex].erase(adjecency_list[fixedVertex].begin() + i);
        size_t j = 0;
        for (; j < adjecency_list[ends[1]].size(); j++) {
          if (adjecency_list[ends[1]][j].first == fixedVertex) {
            break;
          }
        }
        assert(j < adjecency_list[ends[1]].size());
        adjecency_list[ends[1]].erase(adjecency_list[ends[1]].begin() + j);
        degree[ends[1]]--;
        assert(check_degrees());
        start = std::clock();
        recursion_count = 0;
        next_check = 1000;
        find_cycles_from_ends();
        degree[ends[1]]++;
        adjecency_list[ends[1]].insert(adjecency_list[ends[1]].begin() + j, std::make_pair(fixedVertex, e));
        adjecency_list[fixedVertex].insert(adjecency_list[fixedVertex].begin() + i, std::make_pair(ends[1], e));
        vertexStates[ends[1]] = state;
        cycle.clear();
      }
    }
  }

  template<typename G>
  bool EnumerationSolver<G>::find_cycles_from_ends() {
    if (time_limit_reached) {
      return false;
    }
    recursion_count++;
    if (recursion_count >= next_check) {
      //std::cout << recursion_count << std::endl;
      auto time = std::clock();
      if (time > this->max_end_time) {
        this->time_limit_reached = true;
        return false;
      }
      //calculate next_check
      size_t target = CLOCKS_PER_SEC; //we target to check the time every second

      next_check = recursion_count + (recursion_count * target) / (time - start);
    }
    size_t end_index = 0;
    auto end = ends[0];
    std::vector<std::pair<typename G::vertex_descriptor,
        std::pair<typename G::vertex_descriptor, typename G::EdgeId>>> removed_edges;
    std::vector<std::pair<typename G::vertex_descriptor, EnumerationInternal::VertexState>> dropped_vertices;
    std::vector<typename G::vertex_descriptor> fixed_vertices;
    assert(check_degrees());
    assert(check_unsatisfied_edges());
    if (degree[ends[1]] < degree[ends[0]]) {
      end_index = 1;
      end = ends[1];
    }
    typename G::EdgeId possible_cycle_edge_id = -1;
    bool foundLargerCycles = false;
    bool nonClosingEdgeExists = false;
    for (size_t i = 0; i < adjecency_list[end].size(); i++) {
      if (time_limit_reached) {
        //we still want to clean everything up
        break;
      }
      auto next_end = adjecency_list[end][i].first;
      if (vertexStates[next_end] != EnumerationInternal::VertexState::Visited &&
          vertexStates[next_end] != EnumerationInternal::VertexState::Dropped) {
        if (next_end == ends[1 - end_index]) {
          possible_cycle_edge_id = adjecency_list[end][i].second;
        } else {
          bool infeasible = false;
          if (!nonClosingEdgeExists) {
            assert(check_unsatisfied_edges());
            assert(unsatisfied_edges >= 0);
            nonClosingEdgeExists = true;
            //reduce the degrees of all non visited neighbors by 1
            for (size_t j = 0; j < adjecency_list[end].size(); j++) {
              auto v = adjecency_list[end][j].first;
              if (vertexStates[v] != EnumerationInternal::VertexState::Visited) {
                if (vertexStates[v] != EnumerationInternal::VertexState::Dropped) {
                  degree[v]--;
                }
                unsatisfied_edges--;
              }
            }
            vertexStates[end] = EnumerationInternal::VertexState::Visited;
            assert(check_unsatisfied_edges());
          }
          auto next_end_state = vertexStates[next_end];
          vertexStates[next_end] = EnumerationInternal::VertexState::End;
          assert(removed_edges.empty());
          assert(dropped_vertices.empty());
          assert(fixed_vertices.empty());
          //drop degree 1 one neighbors which are not the i-th neighbor.
          for (size_t j = 0; j < adjecency_list[end].size(); j++) {
            if (infeasible) {
              break;
            }
            auto current = adjecency_list[end][j].first;
            if (i != j && vertexStates[current] != EnumerationInternal::VertexState::Dropped &&
                vertexStates[current] != EnumerationInternal::VertexState::Visited) {
              assert(check_degrees());
              while (degree[current] <= 1) {
                if (vertexStates[current] == EnumerationInternal::VertexState::Fixed) {
                  assert(this->minimal);
                  infeasible = true;
                  break;
                } else if (vertexStates[current] == EnumerationInternal::VertexState::End) {
                  if (degree[current] == 0) {
                    infeasible = true;
                  }
                  break;
                } else if (degree[current] == 1) {
                  dropped_vertices.push_back(std::make_pair(current, vertexStates[current]));
                  vertexStates[current] = EnumerationInternal::VertexState::Dropped;
                  for (size_t k = 0; k < adjecency_list[current].size(); k++) {
                    auto next = adjecency_list[current][k].first;
                    if (vertexStates[next] != EnumerationInternal::VertexState::Dropped &&
                        vertexStates[next] != EnumerationInternal::VertexState::Visited) {
                      degree[next]--;
                      current = next;
                      if (this->minimal && vertexStates[next] == EnumerationInternal::VertexState::Unvisited) {
                        vertexStates[next] = EnumerationInternal::VertexState::Fixed;
                        fixed_vertices.push_back(next);
                      }
                      break;
                    }
                  }
                } else {
                  assert(degree[current] == 0);
                  //can only be if current is a neighbor of end
#ifndef  NDEBUG
                  bool neighbor = false;
                  for (size_t k = 0; k < adjecency_list[end].size(); k++) {
                    if (adjecency_list[end][k].first == current) {
                      neighbor = true;
                      break;
                    }
                  }
                  assert(neighbor);
# endif
                  dropped_vertices.push_back(std::make_pair(current, vertexStates[current]));
                  vertexStates[current] = EnumerationInternal::VertexState::Dropped;
                  break;
                }
              }
              assert(check_degrees());
            }
          }
          if (!infeasible) {
            cycle.insert(adjecency_list[end][i].second);
            ends[end_index] = next_end;
            foundLargerCycles = find_cycles_from_ends() || foundLargerCycles;
            ends[end_index] = end;
            assert(cycle.size() > 0);
            cycle.erase(adjecency_list[end][i].second);
          }
          for (auto it = dropped_vertices.rbegin(); it != dropped_vertices.rend(); ++it) {
            auto v = it->first;
            assert(vertexStates[v] == EnumerationInternal::VertexState::Dropped);
            vertexStates[v] = it->second;
            //change degrees
            for (size_t k = 0; k < adjecency_list[v].size(); k++) {
              auto n = adjecency_list[v][k].first;
              if (vertexStates[n] != EnumerationInternal::VertexState::Dropped &&
                  vertexStates[n] != EnumerationInternal::VertexState::Visited) {
                degree[n]++;
                break;
              }
            }
          }
          for (auto& v: fixed_vertices) {
            vertexStates[v] = EnumerationInternal::VertexState::Unvisited;
          }
          dropped_vertices.clear();
          fixed_vertices.clear();
          vertexStates[next_end] = next_end_state;
        }
      }
    }
    if (nonClosingEdgeExists) {
      //increase the degrees of all non visited neighbors by 1

      for (size_t j = 0; j < adjecency_list[end].size(); j++) {
        auto v = adjecency_list[end][j].first;
        if (vertexStates[v] != EnumerationInternal::VertexState::Visited) {
          if (vertexStates[v] != EnumerationInternal::VertexState::Dropped) {
            degree[v]++;
          }
          unsatisfied_edges++;
        }
      }
      vertexStates[end] = EnumerationInternal::VertexState::End;
      assert(check_unsatisfied_edges());
    }
    assert(check_unsatisfied_edges());
    if (possible_cycle_edge_id != (typename G::EdgeId) (-1) && cycle.size() > 1) {
      int satisfied_edges_by_ends = 0;
      for (int i = 0; i < 2; i++) {
        auto _end = ends[i];
        for (auto& p: adjecency_list[_end]) {
          if (vertexStates[p.first] != EnumerationInternal::VertexState::Visited &&
              (i == 0 || vertexStates[p.first] != EnumerationInternal::VertexState::End)) {
            satisfied_edges_by_ends++;
          }
        }
      }
      assert(unsatisfied_edges >= satisfied_edges_by_ends);
      bool dominating = false;
      if (unsatisfied_edges == satisfied_edges_by_ends) {
        dominating = true;
      }
#ifdef  NDEBUG
      cycle.insert(possible_cycle_edge_id);
      Cycle<G> c2(_g, cycle);
      assert(dominating == cycle_is_dominating(c2));
      cycle.erase(possible_cycle_edge_id);
# endif
      if (!this->only_dominating || dominating) {
        //TODO improve this if foundLargerCycle is true => then we know that this cycle can only be interesting if the
        // fixed edge is the edge with the id possible_cycle_edge_id
        cycle.insert(possible_cycle_edge_id);
        Cycle<G> c(_g, cycle);
        assert(dominating == cycle_is_dominating(c));
        if (!this->only_dominating || dominating) {
          this->foundCycle(c);
        }
        cycle.erase(possible_cycle_edge_id);
        return true;
      }
    }

    return foundLargerCycles;
  }
}
#endif //UMC_NEW_ENUMERATIONSOLVER_H
