/*
 * Hasse.h
 *
 *  Created on: Jul 28, 2015
 *      Author: klocker
 */

#ifndef SRC_HASSE_H_
#define SRC_HASSE_H_

#include <functional>
#include <unordered_map>
#include <unordered_set>
#include <cassert>
#include "global.h"

namespace UMC {
  template<class Value, class ValueHasher = ::std::hash<Value>>
  class Hasse {
    public:
      struct Entry {
          explicit Entry(Value val) : val(val) {}

          Value val;
          std::unordered_set<Entry*> children;
          std::unordered_set<Entry*> parents;

          void searchParents(Entry* entry, std::unordered_set<Entry*>& alreadyChecked);

          void searchChildren(Entry* entry, std::unordered_set<Entry*>& alreadyChecked);

          void getAllGreater(std::unordered_set<Value, ValueHasher>& toFill);

          void getAllSmaller(std::unordered_set<Value, ValueHasher>& toFill);
      };

      explicit Hasse(Value root) : root_(new Entry(root)) {
        auto p = entries_.insert(std::make_pair(root, root_));
        assert(p.second);
        _unused(p);
      }

      virtual ~Hasse();

      bool add(const Value& val);

      bool erase(const Value& val);

      const Entry& operator[](const Value& val) { return *entries_.at(val); }

      std::unordered_set<Value, ValueHasher> getAllGreaterThan(const Value& val);

      std::unordered_set<Value, ValueHasher> getAllSmallerThan(const Value& val);

      size_t numEntries() const {
        return entries_.size();
      }

    private:
      std::unordered_map<Value, Entry*, ValueHasher> entries_;
      Entry* root_;
  };

  template<class Value, class ValueHasher>
  Hasse<Value, ValueHasher>::~Hasse() {
    for (auto& p: entries_) {
      delete (p.second);
    }
  }

  template<class Value, class ValueHasher>
  void Hasse<Value, ValueHasher>::Entry::searchParents(UMC::Hasse<Value, ValueHasher>::Entry* entry,
                                                       std::unordered_set<UMC::Hasse<Value, ValueHasher>::Entry*>& alreadyChecked) {
    bool found = false;
    for (Entry* e : this->children) {
      if (e->val < entry->val) {
        found = true;
        if (alreadyChecked.insert(e).second) {
          e->searchParents(entry, alreadyChecked);
        }
      }
    }
    if (!found) {
      //add this to parents
      entry->parents.insert(this);
    }
  }

  template<class Value, class ValueHasher>
  void Hasse<Value, ValueHasher>::Entry::searchChildren(UMC::Hasse<Value, ValueHasher>::Entry* entry,
                                                        std::unordered_set<UMC::Hasse<Value, ValueHasher>::Entry*>& alreadyChecked) {
    if (entry->val < this->val) {
      entry->children.insert(this);
      return;
    }
    for (Entry* e : this->children) {
      if (alreadyChecked.insert(e).second) {
        e->searchChildren(entry, alreadyChecked);
      }
    }
  }

  template<class Value, class ValueHasher>
  bool Hasse<Value, ValueHasher>::add(const Value& val) {
    auto* entry = new Entry(val);
    auto pair = entries_.insert(std::make_pair(val, entry));
    if (!pair.second) {
      delete (entry);
      return false;
    }
    std::unordered_set<Entry*> alreadyChecked;
    root_->searchParents(entry, alreadyChecked);

    assert(entry->parents.size() > 0);

    Entry* firstParent = *entry->parents.begin();

    alreadyChecked.clear();
    firstParent->searchChildren(entry, alreadyChecked);

    for (Entry* e : entry->parents) {
      e->children.insert(entry);
    }

    for (Entry* e : entry->children) {
      for (Entry* p : entry->parents) {
        if (e->parents.erase(p) == 1) {
          p->children.erase(e);
        }
      }
      e->parents.insert(entry);
    }

    return true;
  }

  template<class Value, class ValueHasher>
  bool Hasse<Value, ValueHasher>::erase(const Value& val) {
    if (entries_.count(val) == 0) {
      return false;
    }

    Entry* entry = entries_.at(val);

    for (Entry* child : entry->children) {
      child->parents.erase(entry);
    }
    for (Entry* parent : entry->parents) {
      parent->children.erase(entry);
    }

    for (Entry* child : entry->children) {
      for (Entry* parent : entry->parents) {
        bool hasAlreadyAConnection = false;
        for (Entry* child2 : parent->children) {
          if (child2->val < child->val) {
            hasAlreadyAConnection = true;
            break;
          }
        }
        if (!hasAlreadyAConnection) {
          parent->children.insert(child);
          child->parents.insert(parent);
        }
      }
    }

    entries_.erase(val);
    delete (entry);
    return true;
  }

  template<class Value, class ValueHasher>
  void Hasse<Value, ValueHasher>::Entry::getAllGreater(std::unordered_set<Value, ValueHasher>& toFill) {
    for (Entry* entry : this->children) {
      if (toFill.count(entry->val) == 0) {
        toFill.insert(entry->val);
        entry->getAllGreater(toFill);
      }
    }
  }

  template<class Value, class ValueHasher>
  void Hasse<Value, ValueHasher>::Entry::getAllSmaller(std::unordered_set<Value, ValueHasher>& toFill) {
    for (Entry* entry : this->parents) {
      if (toFill.count(entry->val) == 0) {
        toFill.insert(entry->val);
        entry->getAllSmaller(toFill);
      }
    }
  }

  template<class Value, class ValueHasher>
  std::unordered_set<Value, ValueHasher> Hasse<Value, ValueHasher>::getAllGreaterThan(const Value& val) {
    std::unordered_set<Value, ValueHasher> res;
    entries_.at(val)->getAllGreater(res);
    return res;
  }

  template<class Value, class ValueHasher>
  std::unordered_set<Value, ValueHasher> Hasse<Value, ValueHasher>::getAllSmallerThan(const Value& val) {
    std::unordered_set<Value, ValueHasher> res;
    entries_.at(val)->getAllSmaller(res);
    return res;
  }
} /* namespace UMC */

#endif /* SRC_HASSE_H_ */
