//
// Created by benedikt on 4/12/19.
//

#ifndef UMC_NEW_ILPSOLVER_H
#define UMC_NEW_ILPSOLVER_H

#include "UMCAbstract.h"
#include <gurobi_c++.h>
#include <numeric>
#include <algorithm>
#include "Graph.h"
#include <boost/graph/push_relabel_max_flow.hpp>
#include <limits>
#include <boost/graph/write_dimacs.hpp>
#include <boost/graph/read_dimacs.hpp>

namespace UMC {

  namespace intern {
    template<typename G>
    typename G::vertex_descriptor source(typename G::edge_descriptor e, const G& g) {
      return std::min(boost::source(e, g), boost::target(e, g));
    }

    template<typename G>
    typename G::vertex_descriptor target(typename G::edge_descriptor e, const G& g) {
      return std::max(boost::source(e, g), boost::target(e, g));
    }
  }

  template<typename G>
  class CutCallback;

  enum ILPType {
      IMMEDIATE_CHECK, IMMEDIATE_CHECK_REUSE_CYCLES, MIXED_EDGES_TWO_PHASE, MIXED_EDGES_MIXED_PHASES
  };

  template<typename G>
  class ILPSolver : public UMCAbstract<G> {
    public:
      ILPSolver(ILPType type, bool use_cuts, bool directed = true, bool use_unique_obj = false,
                bool only_dominating = false, bool minimal = false)
          : UMCAbstract<G>(only_dominating, minimal), type(type), use_cuts(use_cuts), directed(directed),
            use_unique_obj(use_unique_obj) {}

      virtual FECycle<G>* searchCycles(const G& g, const std::unordered_set<typename G::EdgeId>& possibleFixedEdges);

      const Cycle<G>* createCycle(const G& g, const std::map<typename G::EdgeId, GRBVar>& y,
                                  bool* bug = nullptr,
                                  CutCallback<G>* callback = nullptr) const;

      FECycle<G>* handleNewFECycle(const G& g, const typename G::EdgeId& fixed_edge, const Cycle<G>* cycle);

      GRBTempConstr constructAvoidCycleConstraint(const Cycle<G>& cycle, const std::vector<GRBVar>& x);

      virtual ~ILPSolver() = default;

      std::vector<GRBTempConstr> getAvoidCycleConstraints(const G& g, const Cycle<G> cycle,
                                                          const std::vector<GRBVar>& x,
                                                          const std::map<typename G::EdgeId, GRBVar>& y,
                                                          const std::map<typename G::EdgeId, GRBVar>& z,
                                                          const typename G::EdgeId* fixed_edge);

    public:
      ILPType type;
      bool use_cuts;
    private:
      bool directed;
      bool use_unique_obj;
      std::unordered_map<typename G::EdgeId, std::vector<Cycle<G>>> cycles;

      const Cycle<G>* searchNewLongestFECycle(const G& g, typename G::EdgeId fixed_edge,
                                              std::vector<Cycle<G>> otherCycles);

      const Cycle<G>* searchNewLongestCycle(const G& g, typename G::vertex_descriptor fixed_vertex,
                                            std::unordered_set<typename G::vertex_descriptor>& forbidden_vertices,
                                            const std::unordered_set<typename G::EdgeId>& possibleFixedEdges);

      const Cycle<G>* searchDifferentCycle(const G& g, typename G::EdgeId fixed_edge,
                                           const Cycle<G>& cycle);

      const Cycle<G>* solveMIPGeneral(const G& g, typename G::vertex_descriptor fixed_vertex,
                                      typename G::EdgeId* fixed_edge, const Cycle<G>* cycle,
                                      const std::vector<Cycle<G>>* otherCycles,
                                      std::unordered_set<typename G::vertex_descriptor>* forbidden_vertices,
                                      const std::unordered_set<typename G::EdgeId>& possibleFixedEdges);

    protected:
      void searchCyclesWithFixedVertex(const G& g, typename G::vertex_descriptor fixedVertex,
                                       std::unordered_set<typename G::vertex_descriptor>& forbidden_vertices,
                                       const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) override;
  };

  template<typename G>
  std::vector<GRBTempConstr> ILPSolver<G>::getAvoidCycleConstraints(const G& g, const Cycle<G> cycle,
                                                                    const std::vector<GRBVar>& x,
                                                                    const std::map<typename G::EdgeId, GRBVar>& y,
                                                                    const std::map<typename G::EdgeId, GRBVar>& z,
                                                                    const typename G::EdgeId* fixed_edge) {
    std::vector<GRBTempConstr> res;
    if (type == IMMEDIATE_CHECK || type == IMMEDIATE_CHECK_REUSE_CYCLES) {
      assert(fixed_edge != nullptr);
      res.push_back(constructAvoidCycleConstraint(cycle, x));
      auto fe_cycle = handleNewFECycle(g, *fixed_edge, &cycle);
      if (fe_cycle != nullptr) { //ignore fe_cycle
        delete fe_cycle;
      }
      if (this->external_bug || this->time_limit_reached()) {
        return res;
      }
    } else {
      assert(type == MIXED_EDGES_TWO_PHASE || type == MIXED_EDGES_MIXED_PHASES);
      this->foundCycle(cycle);
      auto counts = this->db->getCountsOfCycle(cycle);
      GRBLinExpr x_sum;
      for (auto& v: cycle.get_complement_vertices()) {
        x_sum += x.at(v);
      }
      auto m = boost::num_edges(g);
      bool edge_complement_forbidden = false;
      for (auto& id: cycle.get_edges()) {
        if ((counts != nullptr && counts->count(id) > 0 && counts->at(id) == 1) ||
            (fixed_edge != nullptr && id == *fixed_edge) ||
            (counts != nullptr && counts->count(id) > 0 && counts->at(id) == 2 && type == MIXED_EDGES_MIXED_PHASES)) {
          if (type == MIXED_EDGES_TWO_PHASE || (counts != nullptr && counts->count(id) > 0 && counts->at(id) > 1) ||
              (counts == nullptr && fixed_edge != nullptr && id == *fixed_edge)) {
            res.push_back(x_sum >= z.at(id));
          } else {
            assert(
                type == MIXED_EDGES_MIXED_PHASES && counts != nullptr && counts->count(id) > 0 && counts->at(id) == 1);
            GRBLinExpr cycle_x_sum;
            for (auto v : boost::make_iterator_range(boost::vertices(g))) {
              if (!cycle.get_complement_indicator().at(v)) {
                cycle_x_sum += x.at(v);
              }
            }
            auto ce = cycle.get_edges().size();
            res.push_back(ce * x_sum + cycle_x_sum >= ce * z.at(id));
            if (!edge_complement_forbidden) {
              GRBLinExpr edge_complement;
              for (auto p2: y) {
                if (cycle.get_edges().count(p2.first % m) == 0 && cycle.get_edges().count((p2.first % m) + m) == 0) {
                  edge_complement += p2.second;
                }
              }
              for (auto p2: z) {
                if (cycle.get_edges().count(p2.first) == 0 && cycle.get_edges().count(p2.first + m) == 0) {
                  edge_complement += p2.second;
                }
              }
              res.push_back(edge_complement >= 2);
              edge_complement_forbidden = true;
            }
          }
        }
      }
    }
    return res;
  }


  template<typename G>
  GRBTempConstr ILPSolver<G>::constructAvoidCycleConstraint(const Cycle<G>& c, const std::vector<GRBVar>& x) {
    GRBLinExpr vertexComplement;
    for (auto& v: c.get_complement_vertices()) {
      vertexComplement += x.at(v);
    }
    return vertexComplement >= 1;
  }

  template<typename G>
  class CutCallback : public GRBCallback {
      friend ILPSolver<G>;
    public:
      CutCallback(const std::vector<GRBVar>* x, const std::map<typename G::EdgeId, GRBVar>* y,
                  const std::map<typename G::EdgeId, GRBVar>* z, const G* g,
                  bool directed, const Cycle<G>* cycle, typename G::EdgeId* fixed,
                  typename G::vertex_descriptor fixed_vertex, ILPSolver<G>* solver);

    protected:
      void callback() override;

    private:
      typedef boost::adjacency_list_traits<boost::vecS, boost::vecS, boost::directedS> Traits;
      typedef Graph<boost::vecS, boost::vecS, boost::directedS,
          boost::no_property,
          boost::property<boost::edge_capacity_t, long,
              boost::property<boost::edge_residual_capacity_t, long,
                  boost::property<boost::edge_reverse_t, Traits::edge_descriptor> > >
      > DirectedGraph;
      const std::vector<GRBVar>* x;
      const std::map<typename G::EdgeId, GRBVar>* y;
      const std::map<typename G::EdgeId, GRBVar>* z;
      const G* g;
      bool directed;
      const Cycle<G>* cycle;
      DirectedGraph g_dir;
      typename G::EdgeId* fixed;
      typename G::vertex_descriptor fixed_vertex;
      ILPSolver<G>* solver;

      void visit_connected_residual(
          const boost::property_map<DirectedGraph, boost::edge_residual_capacity_t>::type& residual_capacity,
          std::vector<bool>& visited, DirectedGraph::vertex_descriptor v);

      void add_cut(const std::vector<bool>& visited, typename G::vertex_descriptor target);
  };

  template<typename G>
  void CutCallback<G>::visit_connected_residual(
      const boost::property_map<DirectedGraph, boost::edge_residual_capacity_t>::type& residual_capacity,
      std::vector<bool>& visited, DirectedGraph::vertex_descriptor v) {
    visited.at(v) = true;
    //boost::print_graph(g_dir);
    //boost::property_map<DirectedGraph, boost::edge_capacity_t>::type capacity = get(boost::edge_capacity, g_dir);
    for (auto e: boost::make_iterator_range(out_edges(v, g_dir))) {
      if (residual_capacity[e] > 0) {
        auto v2 = boost::target(e, g_dir);
        if (!visited.at(v2)) {
          visit_connected_residual(residual_capacity, visited, v2);
        }
      }
    }
  }

  template<typename G>
  CutCallback<G>::CutCallback(const std::vector<GRBVar>* x, const std::map<typename G::EdgeId, GRBVar>* y,
                              const std::map<typename G::EdgeId, GRBVar>* z, const G* g,
                              bool directed, const Cycle<G>* cycle, typename G::EdgeId* fixed,
                              typename G::vertex_descriptor fixed_vertex, ILPSolver<G>* solver)
      : x(x), y(y), z(z), g(g), directed(directed), cycle(cycle), fixed(fixed), fixed_vertex(fixed_vertex), solver
      (solver) {
    //create directed cycle
    //boost::property_map<DirectedGraph, boost::edge_capacity_t>::type capacity = get(boost::edge_capacity, g_dir);
    boost::property_map<DirectedGraph, boost::edge_reverse_t>::type rev = get(boost::edge_reverse, g_dir);
    //boost::property_map<DirectedGraph, boost::edge_residual_capacity_t>::type residual_capacity =
    //    get(boost::edge_residual_capacity, g_dir);

    auto n = boost::num_vertices(*g);
    auto m = boost::num_edges(*g);
    typedef typename boost::graph_traits<DirectedGraph>::vertex_descriptor vertex_descriptor;
    typedef typename boost::graph_traits<DirectedGraph>::edge_descriptor edge_descriptor;
    vertex_descriptor vs[255]; //list of vertices, only n <= 255 are really used
    for (size_t i = 0; i < n; i++) {
      vs[i] = add_vertex(g_dir);
    }
    for (int i = 0; i < 4; i++) { //i == 0: undireced edges i == 1: directed duplicates, i == 2: reverse of
      // undirected edge, i == 3: reverse of directed duplicate
      for (size_t id = 0; id < m; id++) {
        auto e = g->get_edge(id);
        auto v1 = intern::source(e, *g);
        auto v2 = intern::target(e, *g);
        edge_descriptor e_dir;
        bool in;
        boost::tie(e_dir, in) = g_dir.add_edge(vs[(i == 0 || i == 3) ? v1 : v2], vs[(i == 0 || i == 3) ? v2 : v1]);
        assert(in);
        if (i >= 2) {
          rev[e_dir] = g_dir.get_edge(id + (i == 3 ? m : 0));
          rev[rev[e_dir]] = e_dir;
        }
      }
    }

    //boost::print_graph(g_dir);
  }

  template<typename G>
  FECycle<G>* ILPSolver<G>::handleNewFECycle(const G& g, const typename G::EdgeId& fixed_edge, const Cycle<G>* cycle) {
    assert(cycle != nullptr);
    if (use_cuts) {
      //we have to use the more advanced data structure to recognize uniquely maximal graphs in this case
      this->foundCycle(*cycle);
    } else {
      Output::local.cycles_found++;
      Output::global.cycles_found++;
    }
    auto c2 = searchDifferentCycle(g, fixed_edge, *cycle);
    if (this->external_bug || this->time_limit_reached()) {
      if (c2 != nullptr) {
        delete c2;
      }
      return nullptr;
    }
    if (c2 == nullptr) {
      auto res = new FECycle<G>(*cycle, fixed_edge);
      return res;
    }

    if (use_cuts) {
      this->foundCycle(*c2);
    } else {
      Output::local.cycles_found++;
      Output::global.cycles_found++;
    }

    if (type == IMMEDIATE_CHECK) {
      cycles[fixed_edge].push_back(*cycle);
    } else {
      assert(type == IMMEDIATE_CHECK_REUSE_CYCLES);
      assert(cycle->get_edges().size() == c2->get_edges().size());
      std::vector<typename G::EdgeId> common_edges;
      bool e_found = false;
      for (auto x : cycle->get_edges()) {
        if (c2->get_edges().count(x) > 0) {
          cycles[x].push_back(*cycle);
          if (x == fixed_edge) {
            e_found = true;
          }
        }
      }
      assert(e_found);
      _unused(e_found);
    }
    delete c2;
    return nullptr;
  }

  template<typename G>
  void ILPSolver<G>::searchCyclesWithFixedVertex(const G& g, typename G::vertex_descriptor fixedVertex,
                                                 std::unordered_set<typename G::vertex_descriptor>& forbidden_vertices,
                                                 const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) {
    const Cycle<G>* cycle = nullptr;
    do {
      cycle = searchNewLongestCycle(g, fixedVertex, forbidden_vertices, possibleFixedEdges);
      if (this->external_bug || this->time_limit_reached()) {
        if (cycle != nullptr) {
          delete cycle;
        }
        return;
      }
      if (cycle != nullptr) {
        assert(!use_cuts);
        this->foundCycle(*cycle);
      }
    } while (cycle != nullptr);
  }

  template<typename G>
  FECycle<G>* ILPSolver<G>::searchCycles(const G& g,
                                         const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) {
    if (possibleFixedEdges.size() == 0) {
      return nullptr;
    }
    if (type == IMMEDIATE_CHECK || type == IMMEDIATE_CHECK_REUSE_CYCLES) {
      for (auto& e: possibleFixedEdges) {
        if (cycles.count(e) == 0) {
          cycles[e] = std::vector<Cycle<G>>();
        }
        const Cycle<G>* cycle = nullptr;
        do {
          cycle = searchNewLongestFECycle(g, e, cycles.at(e));
          if (this->external_bug || this->time_limit_reached()) {
            if (cycle != nullptr) {
              delete cycle;
            }
            return nullptr;
          }
          if (cycle != nullptr) {
            assert(!use_cuts); //all cycles must be found and forbidden in the cuts
            auto fe_cycle = handleNewFECycle(g, e, cycle);
            delete (cycle);
            if (this->external_bug || this->time_limit_reached()) {
              assert(fe_cycle == nullptr);
              return nullptr;
            }
            if (fe_cycle != nullptr) {
              return fe_cycle;
            }
          }
        } while (cycle != nullptr);
      }
    } else if (type == MIXED_EDGES_TWO_PHASE || type == MIXED_EDGES_MIXED_PHASES) {
      //first phase;
      this->searchCyclesByFixingVertices(g, possibleFixedEdges);
      if (this->external_bug || this->time_limit_reached()) {
        return nullptr;
      }

      //second phase
      if (type == MIXED_EDGES_TWO_PHASE) {
        for (auto& p: this->db->getCycles()) {
          for (auto p2: p.second.second) {
            if (p2.second == 1) {
              auto& _cycles = this->db->getEdgeToCycleMap().at(p2.first).at(p.first);
              assert(_cycles.size() == 1);
              auto& cycle = **_cycles.begin();
              auto c2 = searchDifferentCycle(g, p2.first, cycle);
              if (this->external_bug || this->time_limit_reached()) {
                if (c2 != nullptr) {
                  delete (c2);
                }
                return nullptr;
              }
              if (c2 == nullptr) {
                auto res = new FECycle<G>(cycle, p2.first);
                return res;
              } else {
                this->foundCycle(*c2);
                delete (c2);
              }
            }
          }
        }
      }
    }

    return nullptr;
  }


  template<typename G>
  const Cycle<G>* ILPSolver<G>::searchNewLongestFECycle(const G& g, typename G::EdgeId fixed_edge,
                                                        std::vector<Cycle<G>> otherCycles) {
    std::unordered_set<typename G::EdgeId> posFixedEdges{fixed_edge};
    return solveMIPGeneral(g, intern::source(g.get_edge(fixed_edge), g), &fixed_edge, nullptr, &otherCycles, nullptr,
                           posFixedEdges);
  }

  template<typename G>
  const Cycle<G>* ILPSolver<G>::searchNewLongestCycle(
      const G& g, typename G::vertex_descriptor fixed_vertex,
      std::unordered_set<typename G::vertex_descriptor>& forbidden_vertices,
      const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) {
    return solveMIPGeneral(g, fixed_vertex, nullptr, nullptr, nullptr, &forbidden_vertices, possibleFixedEdges);
  }

  template<typename G>
  const Cycle<G>* ILPSolver<G>::searchDifferentCycle(const G& g, typename G::EdgeId fixed_edge,
                                                     const Cycle<G>& cycle) {
    std::unordered_set<typename G::EdgeId> posFixedEdges{fixed_edge};
    return solveMIPGeneral(g, intern::source(g.get_edge(fixed_edge), g), &fixed_edge, &cycle, nullptr, nullptr,
                           posFixedEdges);
  }


  template<typename G>
  void CutCallback<G>::add_cut(const std::vector<bool>& visited, typename G::vertex_descriptor target) {
    assert(!visited.at(target));
    GRBLinExpr cut;
    auto m = boost::num_edges(*g);
    for (auto p: *y) {
      auto e = g->get_edge(p.first % m);
      auto v1 = intern::source(e, *g);
      auto v2 = intern::target(e, *g);
      if (visited.at(v1) != visited.at(v2) && (!directed || (visited.at(v1) == (p.first < m)))) {
        cut += p.second;
      }
    }
    this->addLazy(cut >= (directed ? 1 : 2) * (cycle == nullptr ? x->at(target) : GRBLinExpr(1)));
    Output::local.cut_constraints_added++;
    Output::global.cut_constraints_added++;
  }

  template<typename G>
  void CutCallback<G>::callback() {
    if (where == GRB_CB_MIPSOL) {
      auto s = fixed_vertex;
      auto n = boost::num_vertices(*g);
      auto m = boost::num_edges(*g);
      double* x_values = nullptr;
      if (cycle == nullptr) {
        assert(x->size() == n);
        x_values = getSolution(&((*x).at(0)), n);
      }

      std::vector<bool> visited(n, false);

      uint previous = (uint) -1;
      auto current = s;
      auto edge_index = get(boost::edge_index, *g);
      while (previous == (uint) -1 || current != s) {
        visited.at(current) = true;
        //std::cout << current;
        auto pp = previous;
        for (auto e: boost::make_iterator_range(boost::out_edges(current, *g))) {
          auto v1 = boost::source(e, *g);
          auto v2 = boost::target(e, *g);
          assert(v1 == current);
          if (v2 != previous && (cycle == nullptr || !cycle->get_complement_indicator().at(v2))) {
            auto id = get(edge_index, e);
            if (directed && v1 > v2) {
              id = id + m;
            }
            auto y_value = getSolution(y->at(id));
            if (y_value > 0.5) {
              previous = current;
              current = v2;
              break;
            }
          }
        }
        assert(pp != previous);
        _unused(pp);
      }
      //std::cout << std::endl;
      bool found_cut = false;
      for (size_t i = 0; i < n; i++) {
        double x_value;
        if (cycle != nullptr) {
          x_value = cycle->get_complement_indicator().at(i) ? 0.0 : 1.0;
        } else {
          x_value = x_values[i];
        }
        if (x_value > 0.5 && !visited.at(i)) {
          //found cut
          add_cut(visited, i);
          found_cut = true;
          break;
        }
      }
      if (!found_cut && cycle == nullptr && solver->use_cuts) {
        //found a cycle => handle the cycle and then add a cut to avoid it
        bool bug = false;
        auto c = solver->createCycle(*g, *y, solver->catch_bug ? &bug : nullptr, this);
        if (bug) {
          solver->external_bug = true;
        } else {
          typename G::EdgeId new_fixed;
          auto fixed_edge = fixed;
          if (fixed == nullptr && cycle == nullptr) {
            for (auto p: *z) {
              if (getSolution(p.second) > 0.5) {
                new_fixed = p.first;
                fixed_edge = &new_fixed;
                break;
              }
            }
          }
          auto constraints = solver->getAvoidCycleConstraints(*g, *c, *x, *y, *z, fixed_edge);
          if (!solver->external_bug) {
            for (auto& constraint: constraints) {
              this->addLazy(constraint);
            }
          }
        }
        delete (c);
      }

      if (cycle == nullptr) {
        delete[] x_values;
      }
    } else if (where == GRB_CB_MIPNODE) {
      if (getIntInfo(GRB_CB_MIPNODE_STATUS) == GRB_OPTIMAL) {
        boost::property_map<DirectedGraph, boost::edge_capacity_t>::type capacity = get(boost::edge_capacity, g_dir);
        boost::property_map<DirectedGraph, boost::edge_residual_capacity_t>::type residual_capacity =
            get(boost::edge_residual_capacity, g_dir);
        auto m2 = boost::num_edges(g_dir);
        auto m = boost::num_edges(*g);
        assert(cycle != nullptr || (directed && y->size() == 2 * m) || (!directed && y->size() == m));
        assert(m2 == 4 * m);
        _unused(m2);
        _unused(m);

        /*for (auto& p: *y) {
          std::cout << p.second.get(GRB_StringAttr_VarName) << ":" << getNodeRel(p.second) << std::endl;
        }

        for (auto& v: *x) {
          std::cout << v.get(GRB_StringAttr_VarName) << ":" << getNodeRel(v) << std::endl;
        }*/

        auto n = boost::num_vertices(g_dir);

        long granularity = 100 * n;

        for (const auto& e : boost::make_iterator_range(boost::edges(g_dir))) {
          capacity[e] = 0;
        }
        for (auto& p: *y) {
          auto e = g_dir.get_edge(p.first);
          capacity[e] = (long) (getNodeRel(p.second) * granularity + 0.5);
          //std::cout << boost::source(e, g_dir) << "->" << boost::target(e, g_dir) << ":" << capacity[e] << std::endl;
          residual_capacity[e] = capacity[e];
          if (!this->directed) {
            auto e2 = g_dir.get_edge(p.first + m);
            capacity[e2] = capacity[e];
            residual_capacity[e2] = capacity[e];
          }
        }

        auto s = fixed_vertex;

        std::vector<bool> checked(n, false);
        checked.at(s) = true;
        if (fixed != nullptr) {
          checked.at(boost::target(g->get_edge(*fixed), *g)) = true;
        }

        double min_violation = 0.1;

        double* x_values = nullptr;
        if (cycle == nullptr) {
          assert(x->size() == n);
          x_values = getNodeRel(&((*x).at(0)), (int) n);
        }
        for (size_t t = 0; t < n; t++) {
          if (!checked.at(t)) {
            double x_value = cycle != nullptr ? (cycle->get_complement_indicator().at(t) ? 0.0 : 1.0) : x_values[t];
            if (x_value >= min_violation) {
              auto flow = ((double) boost::push_relabel_max_flow(g_dir, s, t)) / granularity;
              Output::local.max_flows_calculated++;
              Output::global.max_flows_calculated++;
              if (flow + min_violation < x_value * (directed ? 1 : 2)) {
                //found a new cut!
                //find vertices connected to source in residual graph
                std::vector<bool> connected_to_source(n, false);
                visit_connected_residual(residual_capacity, connected_to_source, s);
                add_cut(connected_to_source, t);
                break;
              }
            }
          }
        }

        if (cycle == nullptr) {
          delete[] x_values;
        }
      }
    }
  }

  template<typename G>
  const Cycle<G>* ILPSolver<G>::solveMIPGeneral(const G& g, typename G::vertex_descriptor fixed_vertex,
                                                typename G::EdgeId* fixed_edge, const Cycle<G>* cycle,
                                                const std::vector<Cycle<G>>* otherCycles,
                                                std::unordered_set<typename G::vertex_descriptor>*
                                                forbidden_vertices,
                                                const std::unordered_set<typename G::EdgeId>& possibleFixedEdges) {
    try {
      static GRBEnv env; //use static environment to avoid reprinting license information
      GRBModel model = GRBModel(env);

      auto edge_index = boost::get(boost::edge_index, g);
      size_t m = boost::num_edges(g);
      size_t n = boost::num_vertices(g);
      //size_t n_y = (directed ? 2 : 1) * m;
      std::map<typename G::EdgeId, GRBVar> y;
      std::map<typename G::EdgeId, GRBVar> z;
      for (auto e : boost::make_iterator_range(boost::edges(g))) {
        auto id = get(edge_index, e);
        auto v1 = intern::source(e, g);
        auto v2 = intern::target(e, g);
        if (cycle != nullptr && (cycle->get_complement_vertices().count(v1) > 0 ||
                                 cycle->get_complement_vertices().count(v2) > 0)) {
          //this edge gets ignored
          continue;
        }
        y.insert(std::make_pair(id, model.addVar(0.0, 1.0, 0.0, GRB_BINARY,
                                                 "y_" + std::to_string(v1) + "_" + std::to_string(v2))));
        if (directed) {
          y.insert(std::make_pair(id + m, model.addVar(0.0, 1.0, 0.0, GRB_BINARY,
                                                       "y_" + std::to_string(v2) + "_" + std::to_string(v1))));
        }
      }

      std::vector<GRBVar> x;
      if (cycle == nullptr) {
        x.reserve(n);
        for (auto vd : boost::make_iterator_range(boost::vertices(g))) {
          assert(vd == x.size());
          x.push_back(model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "x_" + std::to_string(vd)));
        }

        if (this->minimal) {
          //force neighbors of large vertices to be in the cycle
          for (auto vd : boost::make_iterator_range(boost::vertices(g))) {
            for (auto e: boost::make_iterator_range(boost::out_edges(vd, g))) {
              auto ne = boost::target(e, g);
              assert(ne != vd);
              auto deg = boost::degree(ne, g);
              if (deg > 3 || (fixed_edge != nullptr &&
                              (ne == boost::target(g.get_edge(*fixed_edge), g)
                               || ne == boost::source(g.get_edge(*fixed_edge), g)))) {
                model.addConstr(x.at(vd) == 1, "force_large_neigbhor_" + std::to_string(vd));
                break;
              }
            }
          }
        }
      }

      for (auto i : boost::make_iterator_range(boost::vertices(g))) {
        if (cycle == nullptr || cycle->get_complement_vertices().count(i) == 0) {
          GRBLinExpr incoming;
          GRBLinExpr outgoing;
          for (auto e: boost::make_iterator_range(boost::out_edges(i, g))) {
            auto id = get(edge_index, e);
            if (y.count(id) > 0) {
              if (directed) {
                bool out = intern::source(e, g) == i;
                incoming += y.at(id + (out ? m : 0));
                outgoing += y.at(id + (out ? 0 : m));
              } else {
                outgoing += y.at(id);
              }
            }
          }
          GRBLinExpr rhs = (directed ? 1 : 2) * ((cycle == nullptr) ? x.at(i) : GRBLinExpr(1));
          model.addConstr(outgoing == rhs, "flow_" + ((directed ? "o_" : "") + std::to_string(i)));
          if (directed) {
            model.addConstr(incoming == rhs, "flow_i_" + std::to_string(i));
          }
        }
      }

      if (cycle == nullptr) {
        model.addConstr(x.at(fixed_vertex) == 1, "fixed_vertex");
      }

      if (forbidden_vertices != nullptr) {
        for (auto v: *forbidden_vertices) {
          model.addConstr(x.at(v) == 0, "forbidden_vertex_" + std::to_string(v));
        }
      }

      if (fixed_edge != nullptr) {
        model.addConstr(y.at(*fixed_edge) == 1, "fixed_edge");
      }

      if (directed) {
        for (size_t i = 0; i < m; i++) {
          if (y.count(i) > 0) {
            model.addConstr(y.at(i) + y.at(i + m) <= 1, "directed_" + std::to_string(i));
          }
        }
      }

      if (cycle == nullptr && otherCycles != nullptr) {
        int count = 0;
        for (auto& c: *otherCycles) {
          model.addConstr(constructAvoidCycleConstraint(c, x), "other_cycles_" + std::to_string(++count));
        }
      }

      if (cycle == nullptr && otherCycles == nullptr) {
        assert(fixed_edge == nullptr);
        for (auto id : possibleFixedEdges) {
          auto e = g.get_edge(id);
          auto v1 = intern::source(e, g);
          auto v2 = intern::target(e, g);
          z.insert(std::make_pair(id, model.addVar(0.0, 1.0, 0.0, GRB_BINARY,
                                                   "z_" + std::to_string(v1) + "_" + std::to_string(v2))));
        }

        /*
        TODO: should we use this: if yes we have to change something in the database storage to search for minimal
        graphs since otherwise it would reveal unique cycles with fixed edges which are incident to not used vertices.
        if (this->minimal) {
          //force neighbors of potential fixed edge vertices
          for (auto vd : boost::make_iterator_range(boost::vertices(g))) {
            GRBLinExpr neighboring_edges;
            bool hasLargeNeighbor = false;
            std::unordered_set<int> ids;
            for (auto e: boost::make_iterator_range(boost::out_edges(vd, g))) {
              auto ne = boost::target(e, g);
              assert(ne != vd);
              if (boost::degree(ne, g) > 3) {
                hasLargeNeighbor = true;
                break;
              }
              for (auto e2: boost::make_iterator_range(boost::out_edges(ne, g))) {
                auto id = get(edge_index, e2);
                if (z.count(id) > 0) {
                  neighboring_edges += z.at(id);
                  assert(ids.insert(id).second);
                }
              }
            }
            if (!hasLargeNeighbor) {
              model.addConstr(x.at(vd) >= neighboring_edges, "force_neighbors_of_active_z_" + vd);
            }
          }
        }*/

        for (auto id : possibleFixedEdges) {
          auto e = g.get_edge(id);
          GRBLinExpr rhs = y.at(id);
          if (directed) {
            rhs += y.at(id + m);
          }
          auto v1 = intern::source(e, g);
          auto v2 = intern::target(e, g);
          model.addConstr(z.at(id) <= rhs, "z_y_binding_" + std::to_string(v1) + "_" + std::to_string(v2));
        }

        GRBLinExpr z_sum;
        for (auto& p: z) {
          z_sum += p.second;
        }
        model.addConstr(z_sum == 1, "active_fixed_edge");

        for (auto id : possibleFixedEdges) {
          auto e = g.get_edge(id);
          auto v1 = intern::source(e, g);
          auto v2 = intern::target(e, g);
          int count = 0;
          if (forbidden_vertices != nullptr &&
              (forbidden_vertices->count(v1) > 0 || forbidden_vertices->count(v2) > 0)) {
            continue;
          }
          for (auto& p: this->db->getEdgeToCycleMap().at(id)) {
            GRBLinExpr x_sum;
            for (auto& v: *p.first.vs) {
              x_sum += x.at(v);
            }
            if (this->type == MIXED_EDGES_TWO_PHASE) {
              model.addConstr(x_sum >= z.at(id), "z_active_" + std::to_string(v1) + "_" + std::to_string(v2) + "_" +
                                                 std::to_string(++count));
            } else {
              assert(this->type == MIXED_EDGES_MIXED_PHASES);
              auto c = this->db->getCycles().at(p.first).second.at(id);
              assert(c != 0);
              if (c != (size_t) -1) {
                if (c > 1) {
                  model.addConstr(x_sum >= z.at(id), "z_active>1_" + std::to_string(v1) + "_" +
                                                     std::to_string(v2) + "_" + std::to_string(++count));
                } else if (c == 1) {
                  assert(p.second.size() == 1);
                  auto cyc = *p.second.begin();
                  GRBLinExpr cycle_x_sum;
                  for (auto v : boost::make_iterator_range(boost::vertices(g))) {
                    if (!(*p.first.indicator).at(v)) {
                      cycle_x_sum += x.at(v);
                    }
                  }
                  auto ce = cyc->get_edges().size();
                  model.addConstr(ce * x_sum + cycle_x_sum >= z.at(id) * ce,
                                  "z_active=1_" + std::to_string(v1) + "_" + std::to_string(v2) + "_" +
                                  std::to_string(++count));
                  GRBLinExpr edge_complement;
                  for (auto p2: y) {
                    if (cyc->get_edges().count(p2.first % m) == 0 && cyc->get_edges().count((p2.first % m) + m) == 0) {
                      edge_complement += p2.second;
                    }
                  }
                  for (auto p2: z) {
                    if (cyc->get_edges().count(p2.first) == 0 && cyc->get_edges().count(p2.first + m) == 0) {
                      edge_complement += p2.second;
                    }
                  }
                  model.addConstr(edge_complement >= 2, "z_active=1_different_edge_" +
                                                        std::to_string(v1) + "_" + std::to_string(v2) + "_" +
                                                        std::to_string(count));
                }
              }
            }
          }
        }
      }

      if (cycle != nullptr) {
        GRBLinExpr edgeComplement;
        for (size_t i = 0; i < m; i++) {
          if (y.count(i) > 0 && cycle->get_edges().count(i) == 0) {
            edgeComplement += y.at(i);
            if (directed) {
              edgeComplement += y.at(i + m);
            }
          }
        }
        model.addConstr(edgeComplement >= 1, "new_cycle");
      }

      if (cycle == nullptr && this->only_dominating) {
        for (auto e : boost::make_iterator_range(boost::edges(g))) {
          auto u = boost::source(e, g);
          auto v = boost::target(e, g);
          model.addConstr(x.at(u) + x.at(v) >= 1, "dominating_edge_" + std::to_string(u) + "_" + std::to_string(v));
        }
      }

      if (cycle == nullptr) {
        model.setObjective(std::accumulate(x.begin(), x.end(), GRBLinExpr()), GRB_MAXIMIZE);
      } else if (use_unique_obj) {
        GRBLinExpr common_edges;
        for (auto& id : cycle->get_edges()) {
          common_edges += y.at(id);
          if (directed) {
            common_edges += y.at(id + m);
          }
        }
        model.setObjective(common_edges, GRB_MAXIMIZE);
      }

      auto cb = CutCallback<G>(&x, &y, &z, &g, directed, cycle, fixed_edge, fixed_vertex, this);

      model.setCallback(&cb);

      auto time = std::clock();
      if (time > this->max_end_time) {
        //time limit reached
        return nullptr;
      };
      model.set(GRB_IntParam_Threads, 1);
      model.set(GRB_IntParam_LazyConstraints, 1);
      model.set(GRB_IntParam_LogToConsole, 0);
      model.set(GRB_IntParam_OutputFlag, 0);
      model.set(GRB_StringParam_LogFile, "gurobi.log");
      model.set(GRB_DoubleParam_TimeLimit, std::max((this->max_end_time - time) / (double) CLOCKS_PER_SEC, 3600.0));
      model.set(GRB_IntParam_Seed, this->seed);
      model.set(GRB_DoubleParam_NodefileStart, 3.5);
//model.set(GRB_StringParam_LogFile, "gurobi.log");

      //model.write("model.lp");

//boost::print_graph(g);

      model.optimize();
      Output::local.ilps_solved++;
      Output::global.ilps_solved++;

      switch (model.get(GRB_IntAttr_Status)) {
      case GRB_INFEASIBLE: {
        return nullptr;
      }
      case GRB_OPTIMAL: {
        //model.write("model.sol");
        if (!this->use_cuts || cycle != nullptr) {
          return createCycle(g, y, this->catch_bug ? &this->external_bug : nullptr);
        }
      }
      case GRB_TIME_LIMIT: {
        if (this->catch_bug) {
          this->external_bug = true;
          return nullptr;
        }
        //else we assume that really the mip was too difficult
        this->time_limit = true;
        return nullptr;
      }
      default: {
        std::cerr << "Unexpected Gurobi Error!" << std::endl;
        exit(1);
      }
      }
    } catch (const GRBException& e) {
      std::cerr << "Error code = " << e.getErrorCode() << std::endl;
      std::cerr << e.getMessage() << std::endl;
      exit(1);
    }
  }

  template<typename G>
  const Cycle<G>* ILPSolver<G>::createCycle(const G& g, const std::map<typename G::EdgeId, GRBVar>& y,
                                            bool* bug,
                                            CutCallback<G>* callback) const {
    size_t m = boost::num_edges(g);
    std::unordered_set<typename G::EdgeId> edges;
    for (auto p: y) {
      auto value = callback == nullptr ? p.second.get(GRB_DoubleAttr_X) : callback->getSolution(p.second);
      if (value > 0.5) {
        edges.insert(p.first % m);
      }
    }
    return new Cycle<G>(&g, edges, bug);
  }

}


#endif //UMC_NEW_ILPSOLVER_H
