//
// Created by benedikt on 4/11/19.
//

#ifndef UMC_NEW_UMCABSTRACT_H
#define UMC_NEW_UMCABSTRACT_H

#include "Cycle.h"
#include "CycleDatabase.h"

namespace UMC {
  enum Algorithm {
      ILP,
      Enumeration
  };

  template<typename Graph>
  class UMCAbstract {
    public:
      FECycle<Graph>* findUMC(const Graph& g,
                              const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges);

      bool external_bug;
      bool catch_bug;
      size_t seed;
      std::clock_t max_end_time;
      bool time_limit;

    protected:
      explicit UMCAbstract(bool only_dominating, bool minimal) : external_bug(false), catch_bug(false), seed(0),
                                                                 max_end_time(std::clock_t(-1)),
                                                                 time_limit(false), db(nullptr),
                                                                 only_dominating(only_dominating), minimal(minimal) {}

      CycleDatabase<Graph>* db;
      bool only_dominating;
      bool minimal;

      void searchCyclesByFixingVertices(const Graph& g, const std::unordered_set<typename Graph::EdgeId>&
      possibleFixedEdges);

      virtual void searchCyclesWithFixedVertex(const Graph& g, typename Graph::vertex_descriptor fixedVertex,
                                               std::unordered_set<typename Graph::vertex_descriptor>& forbidden_vertices,
                                               const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges) = 0;

    public:
      virtual ~UMCAbstract() = default;

    protected:

      virtual FECycle<Graph>* searchCycles(const Graph& g,
                                           const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges) = 0;

      void foundCycle(const Cycle<Graph>& cycle) {
        db->addCycle(cycle);
      }

      bool time_limit_reached() const {
        return time_limit || std::clock() > max_end_time;
      }

      FECycle<Graph>* searchFECycleInDB(const Graph& g) const;
  };

  template<typename Graph>
  FECycle<Graph>* UMCAbstract<Graph>::searchFECycleInDB(const Graph& g) const {
    for (auto& p: db->getCycles()) {
      for (auto p2: p.second.second) {
        if (p2.second == 1) {
          bool invalid = false;
          if (minimal) {
            //check if one of the neighbors of the fixed edge is not used because then this cycle is invalid
            typename Graph::vertex_descriptor vs[2];
            auto e = g.get_edge(p2.first);
            vs[0] = boost::source(e, g);
            vs[1] = boost::target(e, g);

            for (int i = 0; i < 2; i++) {
              if (invalid) {
                break;
              }
              for (auto w: boost::make_iterator_range(boost::adjacent_vertices(vs[i], g))) {
                if ((*p.first.indicator)[w]) {
                  invalid = true;
                  break;
                }
              }
            }
          }
          if (!invalid) {
            //found unique maximal cycle!
            auto& cycles = db->getEdgeToCycleMap().at(p2.first).at(p.first);
            assert(cycles.size() == 1);
            return new FECycle<Graph>(**cycles.begin(), p2.first);
          }
        }
      }
    }
    return nullptr;
  }

  template<typename Graph>
  void UMCAbstract<Graph>::searchCyclesByFixingVertices(
      const Graph& g, const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges) {
    std::unordered_set<typename Graph::vertex_descriptor> forbidden_vertices;
    //TODO document this in the working document / paper

    if (this->minimal && boost::num_edges(g) * 2 > boost::num_vertices(g) * 3) {
      //there exists a large vertex => search for neighbor of a large vertex
      bool found = false;
      for (typename Graph::vertex_descriptor v = 0; v < boost::num_vertices(g); v++) {
        if (this->time_limit_reached()) {
          return;
        }
        if (boost::degree(v, g) > 3) {
          for (auto e: boost::make_iterator_range(boost::out_edges(v, g))) {
            if (this->time_limit_reached()) {
              return;
            }
            auto fixed_vertex = boost::target(e, g);
            assert(fixed_vertex != v);
            _unused(fixed_vertex);
            found = true;
            searchCyclesWithFixedVertex(g, v, forbidden_vertices, possibleFixedEdges);
            if (this->external_bug || this->time_limit_reached()) {
              return;
            }
            break;
          }
          break;
        }
      }
      assert(found);
      _unused(found);
    } else if (this->only_dominating) {
      auto& id = *possibleFixedEdges.begin();
      auto e = g.get_edge(id);
      for (int i = 0; i < 2; i++) {
        if (this->time_limit_reached()) {
          return;
        }
        auto fixed = i == 0 ? boost::source(e, g) : boost::target(e, g);
        searchCyclesWithFixedVertex(g, fixed, forbidden_vertices, possibleFixedEdges);
        if (this->external_bug || this->time_limit_reached()) {
          return;
        }
        forbidden_vertices.insert(fixed);
      }
    } else {
      for (typename Graph::vertex_descriptor v = 0; v < boost::num_vertices(g); v++) {
        if (this->time_limit_reached()) {
          return;
        }
        searchCyclesWithFixedVertex(g, v, forbidden_vertices, possibleFixedEdges);
        if (this->external_bug || this->time_limit_reached()) {
          return;
        }
        forbidden_vertices.insert(v);
      }
    }
  }


  template<typename Graph>
  FECycle<Graph>* UMCAbstract<Graph>::findUMC(const Graph& g,
                                              const std::unordered_set<typename Graph::EdgeId>& possibleFixedEdges) {
    CycleDatabase<Graph> _db(&g, possibleFixedEdges);
    db = &_db;
    auto res = searchCycles(g, possibleFixedEdges);
    if (this->external_bug) {
      assert(res == nullptr);
      db = nullptr;
      return nullptr;
    }
    if (this->time_limit_reached()) {
      return nullptr;
    }
    if (res == nullptr) {
      res = searchFECycleInDB(g);
    }
    db = nullptr;
    return res;
  }
}


#endif //UMC_NEW_UMCABSTRACT_H
