#include <iostream>
#include "planar_code.h"
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/program_options.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/boyer_myrvold_planar_test.hpp>
#include "ILPSolver.h"
#include "EnumerationSolver.h"
#include "Graph.h"
#include <ctime>
#include "boost_algorithms/min_max_degree.h"
#include "boost_algorithms/planar_dual.h"
#include <queue>
#include "boost_algorithms/k_connected.h"
#include "boost_algorithms/printing_graphs.h"
#include "boost_algorithms/parse_graphs.h"
#include <fstream>
#include <limits>


//plantri connector function from allow_one_deg3v.c
extern "C" {
void start_plantri(int argc, char* argv[], void (* callback)(unsigned char[], unsigned int), int _dual);
}

namespace UMC {
  typedef Graph<boost::vecS, boost::vecS, boost::undirectedS> G;

  namespace internal {
    boost::program_options::variables_map* vm;
    size_t max_num_graphs = (size_t) -1;
    std::clock_t checkpoint_clock;
    std::clock_t start;
    double time_limit = (double) (std::numeric_limits<std::clock_t>::max()) / (2 * CLOCKS_PER_SEC);
  }

  UMCAbstract<G>* getMethod() {
    bool onlyDominating = internal::vm->count("only-dominating") > 0 ||
                          internal::vm->count("minimal-counterexample") > 0;
    bool minimal = internal::vm->count("minimal-counterexample") > 0;
    switch ((*internal::vm).at("algorithm").as<Algorithm>()) {
    case ILP: {
      return new ILPSolver<G>(
          (*internal::vm).at("ilp-type").as<ILPType>(),         //ILP type
          internal::vm->count("use-cuts") > 0,               //use cuts
          internal::vm->count("undirected") == 0,            //directed
          internal::vm->count("use-uniqueness-obj") > 0,     //use obj in uniqueness ILPs
          onlyDominating,
          minimal
      );
    }
    case Enumeration: {
      return new EnumerationSolver<G>(onlyDominating, minimal);
    }
    default: {
      std::cerr << "Unknown Algorithm used!";
      exit(1);
    }
    }
  }

  void writeGraph(const G& g) {
    boost::print_graph(g);
  }

  typedef typename boost::graph_traits<G>::vertex_descriptor Vertex;
  typedef typename boost::graph_traits<G>::edge_descriptor Edge;

  bool time_limit_reached() {
    if (!Output::time_limit_reached && (double) (std::clock() - internal::start) >
                                       (internal::time_limit - Output::global.runtime) * (double) CLOCKS_PER_SEC) {
      Output::time_limit_reached = true;
    }
    return Output::time_limit_reached;
  }

  bool satisfies_large_depending_properties(const G& g, bool count_globals = true, Edge* fixed = nullptr) {
    //6. |E(G)| <= n3 + |V(G)| - 1/2 (n3* + c*_odd)
    size_t n3 = 0;
    size_t n3s = 0;
    size_t csodd = 0;
    std::vector<bool> small(num_vertices(g), false);
    std::vector<bool> verySmall(num_vertices(g), false);
    for (auto vd : boost::make_iterator_range(vertices(g))) {
      auto d = out_degree(vd, g);
      if (d == 3) {
        small.at(vd) = true;
        n3++;
      }
    }
    if (fixed != nullptr) {
      auto v1 = boost::source(*fixed, g);
      auto v2 = boost::target(*fixed, g);
      small.at(v1) = false;
      small.at(v2) = false;
    }
    for (auto vd : boost::make_iterator_range(vertices(g))) {
      if (small.at(vd)) {
        bool allNeighborsSmall = true;
        for (auto ed : boost::make_iterator_range(out_edges(vd, g))) {
          auto n = boost::target(ed, g);
          assert(n != vd);
          if (!small.at(n)) {
            allNeighborsSmall = false;
            break;
          }
        }
        if (allNeighborsSmall) {
          verySmall.at(vd) = true;
          n3s++;
        }
      }
    }

    std::vector<bool> discovered(num_vertices(g), false);
    //compute c*_odd with bfs on very small vertices
    for (auto vd : boost::make_iterator_range(vertices(g))) {
      if (verySmall.at(vd) && !discovered.at(vd)) {
        std::queue<typename boost::graph_traits<G>::vertex_descriptor> vs;
        vs.push(vd);
        discovered.at(vd) = true;
        size_t component_size = 0;
        while (!vs.empty()) {
          auto current = vs.front();
          vs.pop();
          component_size++;
          for (auto ed : boost::make_iterator_range(out_edges(current, g))) {
            auto n = boost::target(ed, g);
            assert(n != current);
            if (verySmall.at(n) && !discovered.at(n)) {
              vs.push(n);
              discovered.at(n) = true;
            }
          }
        }
        if (component_size % 2 == 1) {
          csodd++;
        }
      }
    }


    if (2 * num_edges(g) > 2 * (n3 + num_vertices(g)) - n3s - csodd) {
      //property not satisfied
      if (count_globals) Output::graphs_filtered_edge_upperbound++;
      return false;
    }

    //7. no cycle or claw of large vertices
    //BFS through the large vertices (can use the same discovered vector which was only used on small vertices
    // until now)

    std::vector<Vertex> predecessors(num_vertices(g), -1);
    for (auto vd : boost::make_iterator_range(vertices(g))) {
      if (!small.at(vd) && !discovered.at(vd)) {
        std::queue<typename boost::graph_traits<G>::vertex_descriptor> vs;
        vs.push(vd);
        discovered.at(vd) = true;
        while (!vs.empty()) {
          auto current = vs.front();
          vs.pop();
          size_t large_neighbors = 0;
          for (auto ed : boost::make_iterator_range(out_edges(current, g))) {
            auto n = boost::target(ed, g);
            assert(n != current);
            if (!small.at(n)) {
              large_neighbors++;
              if (!discovered.at(n)) {
                vs.push(n);
                discovered.at(n) = true;
                predecessors.at(n) = current;
              } else if (n != predecessors.at(current)) {
                //found cycle
                if (count_globals) Output::graphs_filtered_large_vertex_cycle++;
                return false;
              }
            }
          }
          if (large_neighbors >= 3) {
            //found claw
            if (count_globals) Output::graphs_filtered_large_vertex_claw++;
            return false;
          }
        }
      }
    }
    return true;
  }

  void handleGraph(const G& g) {
    if (internal::vm->count("generate-only") > 0) {
      //do nothing
      Output::graphs_checked++;
      return;
    } else if (internal::vm->count("write-only") > 0) {
      //write only the graph
      Output::graphs_checked++;
      boost::print_graph6(std::cout, g);
      std::cout << std::endl;
      //writeGraph(g);
    } else {
      auto time = std::clock();
      UMC::Output::global.runtime += (time - UMC::internal::start) / (double) CLOCKS_PER_SEC;
      UMC::Output::runtime_umc_check += (time - UMC::internal::checkpoint_clock) / (double) CLOCKS_PER_SEC;
      UMC::internal::start = time;
      UMC::internal::checkpoint_clock = time;
      int num_bug_retries = 2;
      for (int i = 0; i < num_bug_retries; i++) {
        auto possible_fixed_edges = std::unordered_set<typename G::EdgeId>();
        auto edge_index = get(boost::edge_index, g);
        for (auto ed : boost::make_iterator_range(edges(g))) {
          if (internal::vm->count("minimal-counterexample") == 0 ||
              satisfies_large_depending_properties(g, false, &ed)) {
            possible_fixed_edges.insert(get(edge_index, ed));
          }
        }
        if (possible_fixed_edges.size() == 0) {
          Output::graphs_filtered_edge_upperbound_no_fixed_edges++;
          return;
        }
        std::clock_t c_start = std::clock();
        Output::local.reset(i == 0);
        auto method = getMethod();
        method->catch_bug = i < num_bug_retries - 1;
        method->seed = i;
        double remaining_sec = internal::time_limit - Output::global.runtime;
        if (remaining_sec <= 0) {
          //timelimit reached
          if (!time_limit_reached()) {
            //bug
            std::cerr << "Time bug occurred!" << std::endl;
            exit(1);
          }
          return;
        }
        method->max_end_time = c_start + (std::clock_t) (remaining_sec * CLOCKS_PER_SEC + 1);
        auto res = method->findUMC(g, possible_fixed_edges);
        if (method->external_bug) {
          Output::local.bugs_occurred++;
          Output::global.bugs_occurred++;
          delete method;
          time = std::clock();
          UMC::internal::start = time;
          UMC::internal::checkpoint_clock = time;
        } else {
          if (method->time_limit) {
            Output::time_limit_reached = true;
            if (res != nullptr) {
              delete res;
            }
          }
          if (!time_limit_reached()) {
            Output::graphs_checked++;
            bool found = res != nullptr;
            if (found) {
              /*std::cout << "FOUND A UNIQUE MAXIMAL CYCLE!!!" << std::endl;
              std::cout << "GRAPH:" << std::endl;
              writeGraph(g);
              std::cout << "CYCLE:" << std::endl;
              std::cout << *res << std::endl;*/
              Output::unique_fe_cycles_found++;
              delete res;
            }


            Output::local.runtime = (std::clock() - c_start) / (double) CLOCKS_PER_SEC;
            std::cout << Output::graphs_checked << ":";
            boost::print_osparse6(std::cout, g);
            std::cout << ",";
            std::cout << boost::num_vertices(g) << "," << boost::num_edges(g) << ",";
            Output::local.print_compact(std::cout);
            std::cout << ":" << found << std::endl;
          }
          delete method;
          return;
        }
      }
    }
  }

  typedef std::vector<std::vector<boost::graph_traits<G>::edge_descriptor >> Embedding;

  bool triangle_free(const G& g, const Embedding* embedding = nullptr) {
    G dual;
    boost::planar_dual(g, dual, embedding);
    return boost::min_degree(dual) >= 4;
  }

  void filter_and_handle_graph(const G& g) {
    if (internal::vm->count("minimal-counterexample") > 0) {
      //check minimal counter example properties

      bool duals = internal::vm->count("use-duals");
      //1. graph is simple and undirected implicitly by the type G
      //no check done here

      //2. graph is planar should always be true since plantri only produces planar graphs
      Embedding embedding(num_vertices(g));
      assert(boyer_myrvold_planarity_test(boost::boyer_myrvold_params::graph = g,
                                          boost::boyer_myrvold_params::embedding = &embedding.at(0)));

      //3. graph has minimum degree 3 is also handled by plantri internal
      assert(boost::min_degree(g) >= 3);

      //4. graph is 3-connected is handled by plantri internal
      //no check done for 3-connected, but we can check 3-edge-connected
      assert(boost::is_3_edge_connected(g));


      //5. graph is triangle free is handled by plantri internal in the dual case
      if (duals) {
        assert(triangle_free(g, &embedding));
      } else {
        if (!triangle_free(g)) {
          Output::graphs_filtered_triangle_free++;
          return;
        }
      }

      if (!satisfies_large_depending_properties(g)) {
        return;
      }

      //6. check if G is cyclically 4-edge connected
      // Since G has minimum degree 3 and has no triangles we know that there exist for the two components of a
      // cyclical 3-edge-cut vertices which are completely inside the component.

      /*bool res = boost::is_essentially_4_edge_connected(g);
      //bool res = true;
      //bool res2 = boost::is_essentially_k_edge_connected_naive(g, 4);
      bool res2 = res;
      if (res != res2) {
        boost::print_graph(g);
        res2 = boost::is_essentially_k_edge_connected_naive(g, 4);
        res = boost::is_essentially_4_edge_connected(g);
        //bool res = true;
        std::cerr << "ERROR at graph "<< global_count <<"!" << std::endl;
      }
      assert (res == res2);*/
      if (!boost::is_essentially_4_edge_connected(g)) {
        //there exists a cyclically 3-edge-cut
        Output::graphs_filtered_cyclically_4_edge_connected++;
        return;
      }
    }
    if (time_limit_reached()) {
      return;
    }
    auto clock = std::clock();
    UMC::Output::runtime_filter += (clock - UMC::internal::checkpoint_clock) / (double) CLOCKS_PER_SEC;
    UMC::internal::checkpoint_clock = clock;
    handleGraph(g);
    clock = std::clock();
    UMC::Output::runtime_umc_check += (clock - UMC::internal::checkpoint_clock) / (double) CLOCKS_PER_SEC;
    UMC::internal::checkpoint_clock = clock;
  }

  void print_output() {
    auto clock = std::clock();
    UMC::Output::global.runtime += (clock - UMC::internal::start) / (double) CLOCKS_PER_SEC;
    UMC::Output::runtime_generation += (clock - UMC::internal::checkpoint_clock) / (double) CLOCKS_PER_SEC;
    if (UMC::internal::only_graphs) {
      return;
    }

    std::cout << UMC::Output::global;
    std::cout << "Time Limit Reached: " << UMC::Output::time_limit_reached << std::endl;
    std::cout << "Total Runtime Generation: " << UMC::Output::runtime_generation << std::endl;
    std::cout << "Total Runtime Filter: " << UMC::Output::runtime_filter << std::endl;
    std::cout << "Total Runtime UMC-Check: " << UMC::Output::runtime_umc_check << std::endl;
    std::cout << "Graphs checked: " << UMC::Output::graphs_checked << std::endl;
    std::cout << "Unique FE-cycles found: " << UMC::Output::unique_fe_cycles_found << std::endl;
    std::cout << "Filtered Graphs by Triangle Free: " << UMC::Output::graphs_filtered_triangle_free << std::endl;
    std::cout << "Filtered Graphs by Edge Upperbound: " << UMC::Output::graphs_filtered_edge_upperbound << std::endl;
    std::cout << "Filtered Graphs by Large Vertex Cycle: " << UMC::Output::graphs_filtered_large_vertex_cycle
              << std::endl;
    std::cout << "Filtered Graphs by Large Vertex Claw: " << UMC::Output::graphs_filtered_large_vertex_claw
              << std::endl;
    std::cout << "Filtered Graphs by Cyclical 4-edge connectivity: " <<
              UMC::Output::graphs_filtered_cyclically_4_edge_connected << std::endl;
    std::cout << "Filtered Graphs by Edge Upperbound with No Possible Fixed Edges: " <<
              UMC::Output::graphs_filtered_edge_upperbound_no_fixed_edges << std::endl;
  }

  void handleGraphCallback(unsigned char code[], unsigned int maxN) {
    if (UMC::Output::graphs_checked >= UMC::internal::max_num_graphs || UMC::time_limit_reached()) {
      print_output();
      exit(0);
    }
    if (UMC::internal::write_probability < 1.0) {
      if (UMC::internal::udis(UMC::internal::gen) >= UMC::internal::write_probability) {
        return;
      }
    }
    UMC::internal::graphs_filter_checked++;
    if (UMC::internal::graphs_filter_checked > UMC::internal::max_graphs_filter_checked) {
      print_output();
      exit(0);
    }
    auto clock = std::clock();
    UMC::Output::runtime_generation += (clock - UMC::internal::checkpoint_clock) / (double) CLOCKS_PER_SEC;
    UMC::internal::checkpoint_clock = clock;
    G g;

    parse_planar_code(reinterpret_cast<char*>(code), maxN, g);

    filter_and_handle_graph(g);
  }

  void call_plantri() {
    assert(internal::vm->count("plantri-command") > 0);
    std::string cmd = (*internal::vm).at("plantri-command").as<std::string>();
    std::replace(cmd.begin(), cmd.end(), '&', ' ');

    std::vector<std::string> args = boost::program_options::split_unix(cmd);

    for (auto it = args.begin(); it != args.end();) {
      if (*it == "2>") {
        it = args.erase(it);
        std::string fileName = *it;
        it = args.erase(it);
        freopen(fileName.c_str(), "a+", stderr);
      } else {
        ++it;
      }
    }

    auto argc = (int) args.size();
    enum {
        kMaxArgs = 64
    };
    char* argv[kMaxArgs];
    for (int i = 0; i < argc; i++) {
      argv[i] = new char[args[i].size() + 1];
      strcpy(argv[i], args[i].c_str());
    }

    start_plantri(argc, argv, &handleGraphCallback, internal::vm->count("use-duals"));

    for (int i = 0; i < argc; i++) {
      delete[] argv[i];
    }
  }

  std::istream& operator>>(std::istream& in, Algorithm& alg) {
    std::string token;
    in >> token;
    if (token == "ILP")
      alg = Algorithm::ILP;
    else if (token == "Enumeration")
      alg = Algorithm::Enumeration;
    else
      in.setstate(std::ios_base::failbit);
    return in;
  }

  std::istream& operator>>(std::istream& in, ILPType& type) {
    std::string token;
    in >> token;
    if (token == "IMMEDIATE_CHECK")
      type = ILPType::IMMEDIATE_CHECK;
    else if (token == "IMMEDIATE_CHECK_REUSE_CYCLES")
      type = ILPType::IMMEDIATE_CHECK_REUSE_CYCLES;
    else if (token == "MIXED_EDGES_TWO_PHASE")
      type = ILPType::MIXED_EDGES_TWO_PHASE;
    else if (token == "MIXED_EDGES_MIXED_PHASES")
      type = ILPType::MIXED_EDGES_MIXED_PHASES;
    else
      in.setstate(std::ios_base::failbit);
    return in;
  }
}


int main(int argc, const char* argv[]) {
  boost::program_options::options_description desc{"Possible usages:\numc-new -f graphs_file [options]\n"
                                                   "umc-new -p plantri_command [options]\nAllowed options"};
  desc.add_options()
      ("help,h", "Help screen")
      ("write-only", "Write only the graphs without checking them")
      ("plantri-command,p", boost::program_options::value<std::string>(), "The plantri command to use")
      ("graphs-file,f", boost::program_options::value<std::string>(), "The file from where to read the graphs")
      ("undirected,u", "If the undirected version should be used for the ILP (default is directed)")
      ("algorithm,a", boost::program_options::value<UMC::Algorithm>()->default_value(UMC::Algorithm::ILP),
       "The algorithm to use, possible values see enum UMC::Algorithm in UMCAbstract.h, default: ILP")
      ("ilp-type,t", boost::program_options::value<UMC::ILPType>()->default_value(UMC::ILPType::IMMEDIATE_CHECK),
       "The ILP type to use, possible values see enum UMC::ILPType in ILPSolver.h, default: IMMEDIATE_CHECK")
      ("use-uniqueness-obj,o", "If we should use an objective for the ILPs which check the uniqueness of a cycle")
      ("only-dominating,d", "Only search for dominating cycles")
      ("use-cuts,c", "If the ILP should use cuts")
      ("minimal-counterexample,m", "If the search should be for a minimal counter example")
      ("use-duals,du", "If duals should get constructed")
      ("generate-only,g", "Only generate graphs, do nothing with them for testing purposes")
      ("max-num-graphs,n", boost::program_options::value<size_t>(), "The maximal number of graphs to process")
      ("write-probability", boost::program_options::value<double>(), "Probability for writing a graph")
      ("only-graph-output", "Only the graphs get outputted")
      ("seed,s", boost::program_options::value<size_t>(), "Seed for randomness")
      ("d2-prob", boost::program_options::value<double>(), "Probability for degree 2 vertex insertion")
      ("max-graphs-filter", boost::program_options::value<size_t>(), "Maximum number of graphs before filtering")
      ("time-limit,l", boost::program_options::value<size_t>(), "time limit in seconds");
  boost::program_options::positional_options_description pd;

  boost::program_options::variables_map vm;
  UMC::internal::vm = &vm;
  store(boost::program_options::command_line_parser(argc, argv).options(desc).positional(pd).allow_unregistered().run(),
        vm);
  notify(vm);

  //check consistency of parameters
  if (vm.count("plantri-command") > 0 && vm.count("graphs-file") > 0) {
    std::cerr << "You have to use either the parameter plantri-command or graphs-file but not both!" << std::endl;
  }

  if (vm.count("plantri-command") == 0 && vm.count("graphs-file") == 0) {
    std::cerr << "You have to use either the parameter plantri-command or graphs-file!" << std::endl;
  }

  if (vm.count("algorithm") > 0 && vm.at("algorithm").as<UMC::Algorithm>() != UMC::Algorithm::ILP &&
      vm.count("ilp-type") > 0 && !vm.at("ilp-type").defaulted()) {
    std::cerr << "Parameter ilp-type can only be used with algorithm ILP" << std::endl;
    return 1;
  }

  if (vm.count("algorithm") > 0 && vm.at("algorithm").as<UMC::Algorithm>() != UMC::Algorithm::ILP &&
      vm.count("undirected") > 0) {
    std::cerr << "Parameter undirected can only be used with algorithm ILP" << std::endl;
    return 1;
  }

  if (vm.count("algorithm") > 0 && vm.at("algorithm").as<UMC::Algorithm>() != UMC::Algorithm::ILP &&
      vm.count("use-uniqueness-obj") > 0) {
    std::cerr << "Parameter use-uniqueness-obj can only be used with algorithm ILP" << std::endl;
    return 1;
  }

  if (vm.count("algorithm") > 0 && vm.at("algorithm").as<UMC::Algorithm>() != UMC::Algorithm::ILP &&
      vm.count("use-cuts") > 0) {
    std::cerr << "Parameter use-cuts can only be used with algorithm ILP" << std::endl;
    return 1;
  }

  if (vm.count("help")) {
    std::cout << desc << std::endl;
    return 0;
  }

  if (vm.count("d2-prob") > 0) {
    UMC::internal::d2prob = vm.at("d2-prob").as<double>();
  }

  if (vm.count("max-graphs-filter") > 0) {
    UMC::internal::max_graphs_filter_checked = vm.at("max-graphs-filter").as<size_t>();
  }

  UMC::Output::global.runtime = 0;
  UMC::internal::start = std::clock();
  UMC::internal::checkpoint_clock = UMC::internal::start;

  if (vm.count("max-num-graphs") > 0) {
    UMC::internal::max_num_graphs = vm.at("max-num-graphs").as<size_t>();
  }

  if (vm.count("time-limit") > 0) {
    UMC::internal::time_limit = vm.at("time-limit").as<size_t>();
  }

  UMC::internal::only_graphs = vm.count("only-graph-output") > 0;
  if (vm.count("write-probability") > 0) {
    UMC::internal::write_probability = vm.at("write-probability").as<double>();
  }

  if (vm.count("seed") > 0) {
    UMC::internal::gen = std::mt19937(vm.at("seed").as<size_t>());
  }

  if (vm.count("plantri-command") > 0) {
    UMC::call_plantri();
  } else if (vm.count("graphs-file") > 0) {
    std::ifstream file(vm.at("graphs-file").as<std::string>());
    while (!file.eof() && file.peek() != EOF && !UMC::time_limit_reached()) {
      if (UMC::Output::graphs_checked >= UMC::internal::max_num_graphs) {
        break;
      }
      UMC::G g;
      boost::parse6(file, g);
      auto c = file.get();
      assert(c == '\n' || file.eof());
      _unused(c);
      filter_and_handle_graph(g);
    }
  }

  UMC::print_output();

  return 0;
}